from binascii import hexlify, unhexlify
import struct

from . import address
from .backends.jsonrpc import JSONRPCWallet
from . import base58
from . import const
from . import ed25519
from . import numbers
from .transaction import Payment, PaymentManager
from .keccak import keccak_256


class Wallet(object):
    """
    Monero wallet.

    Provides interface to operate on a wallet.

    A wallet consists of :class:`accounts <monero.account.Account>`. Fresh wallets start
    with only one account but you may create more. Although it's possible to combine funds
    from different accounts, or even wallets, in a single transaction, this code closely
    follows the idea of separation introduced in the original wallet software.

    The list of accounts will be initialized under the `accounts` attribute.

    The wallet exposes a number of methods that operate on the default account (of index 0).

    :param backend: a wallet backend
    :param \\**kwargs: arguments to initialize a :class:`JSONRPCWallet <monero.backends.jsonrpc.JSONRPCWallet>`
                        instance if no backend is given
    """

    accounts = None

    def __init__(self, backend=None, **kwargs):
        if backend and len(kwargs):
            raise ValueError("backend already given, other arguments are extraneous")

        self._backend = backend if backend else JSONRPCWallet(**kwargs)
        self.incoming = PaymentManager(0, self._backend, "in")
        self.outgoing = PaymentManager(0, self._backend, "out")
        self.refresh()

    def refresh(self):
        """
        Reloads the wallet and its accounts. By default, this method is called only once,
        on :class:`Wallet` initialization. When the wallet is accessed by multiple clients or
        exists in multiple instances, calling `refresh()` will be necessary to update
        the list of accounts.
        """
        self.accounts = self.accounts or []
        idx = 0
        for _acc in self._backend.accounts():
            _acc.wallet = self
            try:
                if self.accounts[idx]:
                    continue
            except IndexError:
                pass
            self.accounts.append(_acc)
            idx += 1

    def height(self):
        """
        Returns the height of the wallet.

        :rtype: int
        """
        return self._backend.height()

    def spend_key(self):
        """
        Returns private spend key. None if wallet is view-only.

        :rtype: str or None
        """
        key = self._backend.spend_key()
        if key == numbers.EMPTY_KEY:
            return None
        return key

    def view_key(self):
        """
        Returns private view key.

        :rtype: str
        """
        return self._backend.view_key()

    def seed(self):
        """
        Returns word seed.

        :rtype: str
        """
        return self._backend.seed()

    def new_account(self, label=None):
        """
        Creates new account, appends it to the :class:`Wallet`'s account list and returns it.

        :param label: account label as `str`
        :rtype: :class:`Account`
        """
        acc, addr = self._backend.new_account(label=label)
        assert acc.index == len(self.accounts)
        self.accounts.append(acc)
        return acc

    def confirmations(self, txn_or_pmt):
        """
        Returns the number of confirmations for given
        :class:`Transaction <monero.transaction.Transaction>` or
        :class:`Payment <monero.transaction.Payment>` object.

        :rtype: int
        """
        if isinstance(txn_or_pmt, Payment):
            txn = txn_or_pmt.transaction
        else:
            txn = txn_or_pmt
        try:
            return max(0, self.height() - txn.height)
        except TypeError:
            return 0

    def export_outputs(self):
        """
        Exports outputs in hexadecimal format.

        :rtype: str
        """
        return self._backend.export_outputs()

    def import_outputs(self, outputs_hex):
        """
        Imports outputs in hexadecimal format. Returns number of imported outputs.

        :rtype: int

        """
        return self._backend.import_outputs(outputs_hex)

    def export_key_images(self):
        """
        Exports signed key images as a list of dicts.

        :rtype: [dict, dict, ...]
        """
        return self._backend.export_key_images()

    def import_key_images(self, key_images_hex):
        """
        Imports key images from a list of dicts. Returns tuple of (height, spent, unspent).

        :rtype: (int, Decimal, Decimal)

        """
        return self._backend.import_key_images(key_images_hex)

    # Following methods operate on default account (index=0)
    def balances(self):
        """
        Returns a tuple of balance and unlocked balance.

        :rtype: (Decimal, Decimal)
        """
        return self.accounts[0].balances()

    def balance(self, unlocked=False):
        """
        Returns specified balance.

        :param unlocked: if `True`, return the unlocked balance, otherwise return total balance
        :rtype: Decimal
        """
        return self.accounts[0].balance(unlocked=unlocked)

    def address(self):
        """
        Returns wallet's master address.

        :rtype: :class:`Address <monero.address.Address>`
        """
        return self.accounts[0].address()

    def addresses(self):
        """
        Returns all addresses of the default account.

        :rtype: list of :class:`Address <monero.address.Address>` and
                :class:`SubAddress <monero.address.SubAddress>`
        """
        return self.accounts[0].addresses()

    def new_address(self, label=None):
        """
        Creates a new address in the default account.

        :rtype: tuple of subaddress, subaddress index (minor):
                (:class:`SubAddress <monero.address.SubAddress>`, `int`)
        """
        return self.accounts[0].new_address(label=label)

    def get_address(self, major, minor):
        """
        Calculates sub-address for account index (`major`) and address index within
        the account (`minor`).

        :rtype: :class:`BaseAddress <monero.address.BaseAddress>`
        """
        # ensure indexes are within uint32
        if major < 0 or major >= 2 ** 32:
            raise ValueError("major index {} is outside uint32 range".format(major))
        if minor < 0 or minor >= 2 ** 32:
            raise ValueError("minor index {} is outside uint32 range".format(minor))
        master_address = self.address()
        if major == minor == 0:
            return master_address
        master_svk = unhexlify(self.view_key())
        master_psk = unhexlify(self.address().spend_key())
        # m = Hs("SubAddr\0" || master_svk || major || minor)
        hsdata = b"".join(
            [
                b"SubAddr\0",
                master_svk,
                struct.pack("<I", major),
                struct.pack("<I", minor),
            ]
        )
        m = keccak_256(hsdata).digest()
        # D = master_psk + m * B
        D = ed25519.edwards_add(
            master_psk, ed25519.scalarmult_B(ed25519.scalar_reduce(m))
        )
        # C = master_svk * D
        C = ed25519.scalarmult(master_svk, D)
        netbyte = bytearray(
            [const.SUBADDR_NETBYTES[const.NETS.index(master_address.net)]]
        )
        data = netbyte + D + C
        checksum = keccak_256(data).digest()[:4]
        return address.SubAddress(base58.encode(hexlify(data + checksum)))

    def transfer(
        self,
        address,
        amount,
        priority=const.PRIO_NORMAL,
        payment_id=None,
        unlock_time=0,
        relay=True,
    ):
        """
        Sends a transfer from the default account. Returns a list of resulting transactions.

        :param address: destination :class:`Address <monero.address.Address>` or subtype
        :param amount: amount to send
        :param priority: transaction priority, implies fee. The priority can be a number
                    from 1 to 4 (unimportant, normal, elevated, priority) or a constant
                    from `monero.const.PRIO_*`.
        :param payment_id: ID for the payment (must be None if
                    :class:`IntegratedAddress <monero.address.IntegratedAddress>`
                    is used as the destination)
        :param unlock_time: the extra unlock delay
        :param relay: if `True`, the wallet will relay the transaction(s) to the network
                    immediately; when `False`, it will only return the transaction(s)
                    so they might be broadcast later
        :rtype: list of :class:`Transaction <monero.transaction.Transaction>`
        """
        return self.accounts[0].transfer(
            address,
            amount,
            priority=priority,
            payment_id=payment_id,
            unlock_time=unlock_time,
            relay=relay,
        )

    def transfer_multiple(
        self,
        destinations,
        priority=const.PRIO_NORMAL,
        payment_id=None,
        unlock_time=0,
        relay=True,
    ):
        """
        Sends a batch of transfers from the default account. Returns a list of resulting
        transactions and amounts.

        :param destinations: a list of destination and amount pairs: [(address, amount), ...]
        :param priority: transaction priority, implies fee. The priority can be a number
                from 1 to 4 (unimportant, normal, elevated, priority) or a constant
                from `monero.const.PRIO_*`.
        :param payment_id: ID for the payment (must be None if
                :class:`IntegratedAddress <monero.address.IntegratedAddress>`
                is used as a destination)
        :param unlock_time: the extra unlock delay
        :param relay: if `True`, the wallet will relay the transaction(s) to the network
                immediately; when `False`, it will only return the transaction(s)
                so they might be broadcast later
        :rtype: list of transaction and amount pairs:
                [(:class:`Transaction <monero.transaction.Transaction>`, `Decimal`), ...]
        """
        return self.accounts[0].transfer_multiple(
            destinations,
            priority=priority,
            payment_id=payment_id,
            unlock_time=unlock_time,
            relay=relay,
        )

    def sweep_all(
        self,
        address,
        priority=const.PRIO_NORMAL,
        payment_id=None,
        subaddr_indices=None,
        unlock_time=0,
        relay=True,
    ):
        """
        Sends all unlocked balance from the default account to an address.
        Returns a list of resulting transactions.

        :param address: destination :class:`Address <monero.address.Address>` or subtype
        :param priority: transaction priority, implies fee. The priority can be a number
                    from 1 to 4 (unimportant, normal, elevated, priority) or a constant
                    from `monero.const.PRIO_*`.
        :param payment_id: ID for the payment (must be None if
                    :class:`IntegratedAddress <monero.address.IntegratedAddress>`
                    is used as the destination)
        :param subaddr_indices: a sequence of subaddress indices to sweep from. Empty sequence
                    or `None` means sweep all positive balances.
        :param unlock_time: the extra unlock delay
        :param relay: if `True`, the wallet will relay the transaction(s) to the network
                    immediately; when `False`, it will only return the transaction(s)
                    so they might be broadcast later
        :rtype: list of :class:`Transaction <monero.transaction.Transaction>`
        """
        return self.accounts[0].sweep_all(
            address,
            priority=priority,
            payment_id=payment_id,
            subaddr_indices=subaddr_indices,
            unlock_time=unlock_time,
            relay=relay,
        )



#!/usr/bin/env python

# Electrum - lightweight Bitcoin client
# Copyright (C) 2011 thomasv@gitorious
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Note about US patent no 5892470: Here each word does not represent a given digit.
# Instead, the digit represented by a word is variable, it depends on the previous word.
#
# Copied 17 February 2018 from MoneroPy, originally from Electrum:
#     https://github.com/bigreddmachine/MoneroPy/blob/master/moneropy/mnemonic.py ch: 80cc16c39b16c55a8d052fbf7fae68644f7a5f02
#     https://github.com/spesmilo/electrum/blob/master/lib/old_mnemonic.py ch:9a0aa9b4783ea03ea13c6d668e080e0cdf261c5b
#
# Significantly modified on 26 May 2018 by Michal Salaban:
#   + support for 12/13-word seeds
#   + simplified interface, changed exceptions (assertions -> explicit raise)
#   + optimization

from binascii import hexlify, unhexlify
from os import urandom
from . import base58, const, ed25519, wordlists
from .address import address
from .keccak import keccak_256


class Seed(object):
    """Creates a seed object either from local system randomness or an imported phrase.

    :rtype: :class:`Seed <monero.seed.Seed>`
    """

    def __init__(self, phrase_or_hex="", wordlist="English"):
        """If user supplied a seed string to the class, break it down and determine
        if it's hexadecimal or mnemonic word string. Gather the values and store them.
        If no seed is passed, automatically generate a new one from local system randomness.

        :rtype: :class:`Seed <monero.seed.Seed>`
        """
        self.phrase = ""  # 13 or 25 word mnemonic word string
        self.hex = ""  # hexadecimal

        self.word_list = wordlists.get_wordlist(wordlist)

        self._ed_pub_spend_key = None
        self._ed_pub_view_key = None

        if phrase_or_hex:
            seed_split = phrase_or_hex.split(" ")
            if len(seed_split) >= 24:
                # standard mnemonic
                self.phrase = phrase_or_hex
                if len(seed_split) == 25:
                    # with checksum
                    self._validate_checksum()
                self._decode_seed()
            elif len(seed_split) >= 12:
                # mymonero mnemonic
                self.phrase = phrase_or_hex
                if len(seed_split) == 13:
                    # with checksum
                    self._validate_checksum()
                self._decode_seed()
            elif len(seed_split) == 1:
                # single string, probably hex, but confirm
                if not len(phrase_or_hex) % 8 == 0:
                    raise ValueError(
                        "Not valid hexadecimal: {hex}".format(hex=phrase_or_hex)
                    )
                self.hex = phrase_or_hex
                self._encode_seed()
            else:
                raise ValueError(
                    "Not valid mnemonic phrase or hex: {arg}".format(arg=phrase_or_hex)
                )
        else:
            self.hex = generate_random_hex()
            self._encode_seed()

    def is_mymonero(self):
        """Returns True if the seed is MyMonero-style (12/13-word)."""
        return len(self.hex) == 32

    def _encode_seed(self):
        """Convert hexadecimal string to mnemonic word representation with checksum."""
        self.phrase = self.word_list.encode(self.hex)

    def _decode_seed(self):
        """Calculate hexadecimal representation of the phrase."""
        self.hex = self.word_list.decode(self.phrase)

    def _validate_checksum(self):
        """Given a mnemonic word string, confirm seed checksum (last word) matches the computed checksum.

        :rtype: bool
        """
        phrase = self.phrase.split(" ")
        if self.word_list.get_checksum(self.phrase) == phrase[-1]:
            return True
        raise ValueError("Invalid checksum")

    def hex_seed(self):
        return self.hex

    def _hex_seed_keccak(self):
        return keccak_256(unhexlify(self.hex)).digest()

    def secret_spend_key(self):
        a = self._hex_seed_keccak() if self.is_mymonero() else unhexlify(self.hex)
        return hexlify(ed25519.scalar_reduce(a)).decode()

    def secret_view_key(self):
        b = (
            self._hex_seed_keccak()
            if self.is_mymonero()
            else unhexlify(self.secret_spend_key())
        )
        return hexlify(ed25519.scalar_reduce(keccak_256(b).digest())).decode()

    def public_spend_key(self):
        if self._ed_pub_spend_key:
            return self._ed_pub_spend_key
        self._ed_pub_spend_key = ed25519.public_from_secret_hex(self.secret_spend_key())
        return self._ed_pub_spend_key

    def public_view_key(self):
        if self._ed_pub_view_key:
            return self._ed_pub_view_key
        self._ed_pub_view_key = ed25519.public_from_secret_hex(self.secret_view_key())
        return self._ed_pub_view_key

    def public_address(self, net=const.NET_MAIN):
        """Returns the master :class:`Address <monero.address.Address>` represented by the seed.

        :param net: the network, one of `const.NET_*`; default is `const.NET_MAIN`

        :rtype: :class:`Address <monero.address.Address>`
        """
        # backward compatibility
        _net = net[:-3] if net.endswith("net") else net
        if net not in const.NETS:
            raise ValueError(
                "Invalid net argument '{:s}'. Must be one of monero.const.NET_*".format(
                    net
                )
            )
        netbyte = (18, 53, 24)[const.NETS.index(net)]
        data = "{:x}{:s}{:s}".format(
            netbyte, self.public_spend_key(), self.public_view_key()
        )
        checksum = keccak_256(unhexlify(data)).hexdigest()
        return address(base58.encode(data + checksum[0:8]))


def generate_random_hex(n_bytes=32):
    """Generate a secure and random hexadecimal string. 32 bytes by default, but arguments can override.

    :rtype: str
    """
    h = hexlify(urandom(n_bytes))
    return "".join(h.decode("utf-8"))



import binascii
import itertools
import operator
import re
import struct
import varint
import warnings
from ..address import address
from ..numbers import from_atomic, PaymentID
from .. import ed25519
from .. import exceptions
from .extra import ExtraParser
from ..keccak import keccak_256


class Payment(object):
    """
    A payment base class, representing payment not associated with any
    :class:`Account <monero.account.Account>`.

    This class is not intended to be turned into objects by the user,
    it is used by backends.
    """

    payment_id = None
    amount = None
    timestamp = None
    transaction = None
    local_address = None
    note = ""

    _reprstr = "{} @ {} {:.12f} id={}"

    def __init__(self, **kwargs):
        self.amount = kwargs.pop("amount", self.amount)
        self.timestamp = kwargs.pop("timestamp", self.timestamp)
        self.payment_id = kwargs.pop("payment_id", self.payment_id)
        self.transaction = kwargs.pop("transaction", self.transaction)
        self.local_address = kwargs.pop("local_address", self.local_address)
        self.note = kwargs.pop("note", self.note)
        if len(kwargs):
            raise ValueError(
                "Excessive arguments for {}: {}".format(type(self), kwargs)
            )

    def __repr__(self):
        return self._reprstr.format(
            self.transaction.hash,
            self.transaction.height or "pool",
            self.amount,
            self.payment_id,
        )


class IncomingPayment(Payment):
    """
    An incoming payment (one that increases the balance of an
    :class:`Account <monero.account.Account>`)
    """

    _reprstr = "in: {} @ {} {:.12f} id={}"


class OutgoingPayment(Payment):
    """
    An outgoing payment (one that decreases the balance of an
    :class:`Account <monero.account.Account>`)
    """

    destinations = None

    def __init__(self, **kwargs):
        self.destinations = kwargs.pop("destinations", [])
        super(OutgoingPayment, self).__init__(**kwargs)

    _reprstr = "out: {} @ {} {:.12f} id={}"


class Transaction(object):
    """
    A Monero transaction. Identified by `hash`, it can be a part of a block of some `height`
    or not yet mined (`height` is `None` then).

    This class is not intended to be turned into objects by the user,
    it is used by backends.
    """

    hash = None
    fee = None
    height = None
    timestamp = None
    key = None
    blob = None
    confirmations = None
    output_indices = None
    json = None
    version = None
    pubkeys = None

    @property
    def is_coinbase(self):
        if self.json:
            return "gen" in self.json["vin"][0]
        raise exceptions.TransactionWithoutJSON(
            "Tx {:s} has no .json attribute and it cannot be determined "
            "if it's coinbase tx.".format(self.hash)
        )

    @property
    def size(self):
        if not self.blob:
            raise exceptions.TransactionWithoutBlob(
                "Transaction has no blob, hence the size cannot be determined. "
                "Perhaps the backend prunes transaction data?"
            )
        return len(self.blob)

    def __init__(self, **kwargs):
        self.hash = kwargs.get("hash", self.hash)
        self.fee = kwargs.get("fee", self.fee)
        self.height = kwargs.get("height", self.height)
        self.timestamp = kwargs.get("timestamp", self.timestamp)
        self.key = kwargs.get("key", self.key)
        self.blob = kwargs.get("blob", self.blob)
        self.confirmations = kwargs.get("confirmations", self.confirmations)
        self.output_indices = kwargs.get("output_indices", self.output_indices)
        self.json = kwargs.get("json", self.json)
        self.pubkeys = self.pubkeys or []
        if self.json:
            if "rct_signatures" in self.json:
                self.version = 2
            else:
                self.version = 1

    def outputs(self, wallet=None):
        """
        Returns a list of outputs. If wallet is given, decodes destinations and amounts
        for outputs directed to the wallet, provided that matching subaddresses have been
        already generated.
        """

        def _scan_pubkeys(
            svk, psk, stealth_address, amount, encamount, commitment, on_chain_vt
        ):
            for keyidx, tx_key in enumerate(self.pubkeys):
                # precompute
                svk_2 = ed25519.scalar_add(svk, svk)
                svk_4 = ed25519.scalar_add(svk_2, svk_2)
                svk_8 = ed25519.scalar_add(svk_4, svk_4)
                #
                shared_secret = ed25519.scalarmult(svk_8, tx_key)
                if on_chain_vt:
                    vt_hsdata = b"".join(
                        [b"view_tag", shared_secret, varint.encode(idx)]
                    )
                    vt_full = keccak_256(vt_hsdata).digest()
                    vt = vt_full[0:1]
                    if vt != on_chain_vt:
                        # short-circuit so it doesn't have to do the rest of this for ~99.6% of outputs
                        # the view tag check yields false positives 1/256 times, because it's just 1 byte
                        continue

                hsdata = b"".join(
                    [
                        shared_secret,
                        varint.encode(idx),
                    ]
                )
                Hs_ur = keccak_256(hsdata).digest()
                Hs = ed25519.scalar_reduce(Hs_ur)
                k = ed25519.edwards_add(
                    ed25519.scalarmult_B(Hs),
                    psk,
                )
                if k != stealth_address:
                    continue
                if not encamount:
                    # Tx ver 1
                    return Payment(
                        amount=amount,
                        timestamp=self.timestamp,
                        transaction=self,
                        local_address=addr,
                    )
                amount_hs = keccak_256(b"amount" + Hs).digest()
                xormask = amount_hs[: len(encamount)]
                dec_amount = bytearray(
                    a ^ b for a, b in zip(*map(bytearray, (encamount, xormask)))
                )
                # verify that the commitment == yG + bH
                # https://web.getmonero.org/library/Zero-to-Monero-2-0-0.pdf#section.5.3
                y = ed25519.scalar_reduce(keccak_256(b"commitment_mask" + Hs).digest())
                yG = ed25519.scalarmult_B(y)
                b = ed25519.scalar_reduce(bytes(dec_amount))
                bH = ed25519.scalarmult_H(b)
                amount = 0
                if commitment == ed25519.edwards_add(yG, bH):
                    int_amount = struct.unpack("<Q", dec_amount)[0]
                    amount = from_atomic(int_amount)
                    return Payment(
                        amount=amount,
                        timestamp=self.timestamp,
                        transaction=self,
                        local_address=addr,
                    )

        if not self.json:
            raise exceptions.TransactionWithoutJSON(
                "Tx {:s} has no .json attribute".format(self.hash)
            )

        if wallet:
            ep = ExtraParser(self.json["extra"])
            extra = ep.parse()
            self.pubkeys = extra.get("pubkeys", [])
            svk = binascii.unhexlify(wallet.view_key())
            # fetch before loop to save on calls; cast to list to preserve over multiple iterations
            addresses = list(
                itertools.chain(
                    *map(operator.methodcaller("addresses"), wallet.accounts)
                )
            )
        """
        pre hard fork:
        { 
          "target": {
            "key": "ea3f..."
          }
        }
        post hard fork:
        { 
          "target": {
            "tagged_key": {
              "key": "ea3f...",
              "view_tag": "a1"
            }
          }
        }
        """
        outs = []
        for idx, vout in enumerate(self.json["vout"]):
            # see if post hard fork json structure present:
            try:
                if vout["target"]["tagged_key"]:
                    # post fork transaction
                    stealth_address = binascii.unhexlify(
                        vout["target"]["tagged_key"]["key"]
                    )
                    orig_stealth_address = vout["target"]["tagged_key"]["key"]
                    on_chain_vt = binascii.unhexlify(
                        vout["target"]["tagged_key"]["view_tag"]
                    )
            except:
                # pre fork transaction
                stealth_address = binascii.unhexlify(vout["target"]["key"])
                orig_stealth_address = vout["target"]["key"]
                on_chain_vt = False
                pass

            encamount = None
            commitment = None
            if self.version == 2 and not self.is_coinbase:
                encamount = binascii.unhexlify(
                    self.json["rct_signatures"]["ecdhInfo"][idx]["amount"]
                )
                commitment = binascii.unhexlify(
                    self.json["rct_signatures"]["outPk"][idx]
                )
            payment = None
            amount = (
                from_atomic(vout["amount"])
                if self.version == 1 or self.is_coinbase
                else None
            )
            if wallet:
                for addridx, addr in enumerate(addresses):
                    psk = binascii.unhexlify(addr.spend_key())
                    payment = _scan_pubkeys(
                        svk,
                        psk,
                        stealth_address,
                        amount,
                        encamount,
                        commitment,
                        on_chain_vt,
                    )
                    if payment:
                        break
            outs.append(
                Output(
                    stealth_address=orig_stealth_address,
                    amount=payment.amount if payment else amount,
                    index=self.output_indices[idx] if self.output_indices else None,
                    transaction=self,
                    payment=payment,
                )
            )
        return outs

    def __repr__(self):
        return self.hash


class Output(object):
    """
    A Monero one-time public output (A.K.A stealth address).
    Identified by `stealth_address`, or `index` and `amount`
    together, it can contain differing levels of information on an output.

    This class is not intended to be turned into objects by the user,
    it is used by backends.
    """

    stealth_address = None
    amount = None
    index = None
    transaction = None
    payment = None

    def __init__(self, **kwargs):
        self.stealth_address = kwargs.get("stealth_address", self.stealth_address)
        self.amount = kwargs.get("amount", self.amount)
        self.index = kwargs.get("index", self.index)
        self.transaction = kwargs.get("transaction", self.transaction)
        self.payment = kwargs.get("payment", self.payment)

    def __repr__(self):
        # Try to represent output as (index, amount) pair if applicable because there is no RPC
        # daemon command to lookup outputs by their stealth_address ;(
        if self.stealth_address:
            res = self.stealth_address
        else:
            res = "(index={},amount={})".format(self.index, self.amount)
        if self.payment:
            return "{:s}, {:.12f} to [{:s}]".format(
                res, self.payment.amount, str(self.payment.local_address)[:6]
            )
        return res

    def __eq__(self, other):
        # Try to compare stealth_addresses, then try to compare (index,amount) pairs, else raise error
        if self.stealth_address and other.stealth_address:
            return self.stealth_address == other.stealth_address
        elif None not in (self.index, other.index, self.amount, other.amount):
            return self.index == other.index and self.amount == other.amount
        else:
            raise TypeError(
                "Given one-time outputs (%r,%r) are not comparable".format(self, other)
            )

    def __ne__(self, other):
        return not (self == other)


class PaymentManager(object):
    """
    A payment query manager, handling either incoming or outgoing payments of
    an :class:`Account <monero.account.Account>`.

    This class is not intended to be turned into objects by the user,
    it is used by backends.
    """

    account_idx = 0
    backend = None

    def __init__(self, account_idx, backend, direction):
        self.account_idx = account_idx
        self.backend = backend
        self.direction = direction

    def __call__(self, **filterparams):
        fetch = (
            self.backend.transfers_in
            if self.direction == "in"
            else self.backend.transfers_out
        )
        return fetch(self.account_idx, PaymentFilter(**filterparams))


def _validate_tx_id(txid):
    if not bool(re.compile("^[0-9a-f]{64}$").match(txid)):
        raise ValueError(
            "Transaction ID must be a 64-character hexadecimal string, not "
            "'{}'".format(txid)
        )
    return txid


class _ByHeight(object):
    """A helper class used as key in sorting of payments by height.
    Mempool goes on top, blockchain payments are ordered with descending block numbers.

    **WARNING:** Integer sorting is reversed here.
    """

    def __init__(self, pmt):
        self.pmt = pmt

    def _cmp(self, other):
        sh = self.pmt.transaction.height
        oh = other.pmt.transaction.height
        if sh is oh is None:
            return 0
        if sh is None:
            return 1
        if oh is None:
            return -1
        return (sh > oh) - (sh < oh)

    def __lt__(self, other):
        return self._cmp(other) > 0

    def __le__(self, other):
        return self._cmp(other) >= 0

    def __eq__(self, other):
        return self._cmp(other) == 0

    def __ge__(self, other):
        return self._cmp(other) <= 0

    def __gt__(self, other):
        return self._cmp(other) < 0

    def __ne__(self, other):
        return self._cmp(other) != 0


class PaymentFilter(object):
    """
    A helper class that filters payments retrieved by the backend.

    This class is not intended to be turned into objects by the user,
    it is used by backends.
    """

    def __init__(self, **filterparams):
        self.min_height = filterparams.pop("min_height", None)
        self.max_height = filterparams.pop("max_height", None)
        self.unconfirmed = filterparams.pop("unconfirmed", False)
        self.confirmed = filterparams.pop("confirmed", True)
        _local_address = filterparams.pop("local_address", None)
        _tx_id = filterparams.pop("tx_id", None)
        _payment_id = filterparams.pop("payment_id", None)
        if len(filterparams) > 0:
            raise ValueError(
                "Excessive arguments for payment query: {}".format(filterparams)
            )
        if self.unconfirmed and (
            self.min_height is not None or self.max_height is not None
        ):
            warnings.warn(
                "Height filtering (min_height/max_height) has been requested while "
                "also asking for unconfirmed transactions. These are mutually exclusive. "
                "As mempool transactions have no height at all, they will be excluded "
                "from the result.",
                RuntimeWarning,
            )
        if _local_address is None:
            self.local_addresses = []
        else:
            if isinstance(_local_address, str):
                local_addresses = [_local_address]
            else:
                try:
                    iter(_local_address)
                    local_addresses = _local_address
                except TypeError:
                    local_addresses = [_local_address]
            self.local_addresses = list(map(address, local_addresses))
        if _tx_id is None:
            self.tx_ids = []
        else:
            if isinstance(_tx_id, str):
                tx_ids = [_tx_id]
            else:
                try:
                    iter(_tx_id)
                    tx_ids = _tx_id
                except TypeError:
                    tx_ids = [_tx_id]
            self.tx_ids = list(map(_validate_tx_id, tx_ids))
        if _payment_id is None:
            self.payment_ids = []
        else:
            if isinstance(_payment_id, str):
                payment_ids = [_payment_id]
            else:
                try:
                    iter(_payment_id)
                    payment_ids = _payment_id
                except TypeError:
                    payment_ids = [_payment_id]
            self.payment_ids = list(map(PaymentID, payment_ids))

    def check(self, payment):
        ht = payment.transaction.height
        if ht is None:
            if not self.unconfirmed:
                return False
            if self.min_height is not None or self.max_height is not None:
                # mempool txns are filtered out if any height range check is present
                return False
        else:
            if not self.confirmed:
                return False
            if self.min_height is not None and ht < self.min_height:
                return False
            if self.max_height is not None and ht > self.max_height:
                return False
        if self.payment_ids and payment.payment_id not in self.payment_ids:
            return False
        if self.tx_ids and payment.transaction.hash not in self.tx_ids:
            return False
        if self.local_addresses and payment.local_address not in self.local_addresses:
            return False
        return True

    def filter(self, payments):
        return sorted(filter(self.check, payments), key=_ByHeight)



class MoneroException(Exception):
    pass


class BackendException(MoneroException):
    pass


class NoDaemonConnection(BackendException):
    pass


class DaemonIsBusy(BackendException):
    pass


class AccountException(MoneroException):
    pass


class WrongAddress(AccountException):
    pass


class WrongPaymentId(AccountException):
    pass


class NotEnoughMoney(AccountException):
    pass


class NotEnoughUnlockedMoney(NotEnoughMoney):
    pass


class AmountIsZero(AccountException):
    pass


class TransactionNotPossible(AccountException):
    pass


class TransactionBroadcastError(BackendException):
    def __init__(self, message, details=None):
        self.details = details
        super(TransactionBroadcastError, self).__init__(message)


class TransactionNotFound(AccountException):
    pass


class SignatureCheckFailed(MoneroException):
    pass


class WalletIsNotDeterministic(MoneroException):
    pass


class GenericTransferError(AccountException):
    pass


class AccountIndexOutOfBound(AccountException):
    pass


class AddressIndexOutOfBound(AccountException):
    pass


class WalletIsWatchOnly(MoneroException):
    pass


class TransactionIncomplete(MoneroException):
    pass


class TransactionWithoutBlob(TransactionIncomplete):
    pass


class TransactionWithoutJSON(TransactionIncomplete):
    pass



from .. import exceptions
from ..account import Account
from ..address import Address
from ..numbers import EMPTY_KEY
from ..seed import Seed


class WalletIsOffline(exceptions.BackendException):
    pass


class OfflineWallet(object):
    """
    Offline backend for Monero wallet. Provides support for address generation.
    """

    _address = None
    _svk = None
    _ssk = EMPTY_KEY

    def __init__(self, address, view_key=None, spend_key=None):
        self._address = Address(address)
        self._svk = view_key or self._svk
        self._ssk = spend_key or self._ssk

    def height(self):
        raise WalletIsOffline()

    def spend_key(self):
        return self._ssk

    def view_key(self):
        return self._svk

    def seed(self):
        return Seed(self._ssk)

    def accounts(self):
        return [Account(self, 0)]

    def new_account(self, label=None):
        raise WalletIsOffline()

    def addresses(self, account=0, addr_indices=None):
        if account == 0 and (addr_indices == [0] or addr_indices is None):
            return [self._address]
        raise WalletIsOffline()  # pragma: no cover (this should never happen)

    def new_address(self, account=0, label=None):
        raise WalletIsOffline()

    def balances(self, account=0):
        raise WalletIsOffline()

    def transfers_in(self, account, pmtfilter):
        raise WalletIsOffline()

    def transfers_out(self, account, pmtfilter):
        raise WalletIsOffline()

    def export_outputs(self):
        raise WalletIsOffline()

    def import_outputs(self, outputs_hex):
        raise WalletIsOffline()

    def export_key_images(self):
        raise WalletIsOffline()

    def import_key_images(self, key_images):
        raise WalletIsOffline()

    def transfer(self, *args, **kwargs):
        raise WalletIsOffline()



import operator
from .transaction import Transaction


class Block(object):
    """
    A Monero block. Identified by `hash` (optionaly by `height`).

    This class is not intended to be turned into objects by the user,
    it is used by backends.
    """

    hash = None
    height = None
    timestamp = None
    version = None
    difficulty = None
    nonce = None
    orphan = False
    prev_hash = None
    reward = None
    blob = None

    transactions = None

    def __init__(self, **kwargs):
        for k in (
            "hash",
            "height",
            "timestamp",
            "version",
            "difficulty",
            "nonce",
            "prev_hash",
            "reward",
        ):
            setattr(self, k, kwargs.get(k, getattr(self, k)))
        self.orphan = kwargs.get("orphan", self.orphan)
        self.transactions = kwargs.get("transactions", self.transactions or [])
        self.blob = kwargs.get("blob", self.blob)

    def __eq__(self, other):
        if isinstance(other, Block):
            return self.hash == other.hash
        elif isinstance(other, str):
            return self.hash == other
        return super(Block, self).__eq__(other)

    def __contains__(self, tx):
        if isinstance(tx, str):
            txid = tx
        elif isinstance(tx, Transaction):
            txid = tx.hash
        else:
            raise ValueError(
                "Only Transaction or hash strings may be used to test existence in Blocks, "
                "got '{:s}'".format(tx)
            )
        return txid in map(operator.attrgetter("hash"), self.transactions)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from binascii import crc32


WORDLISTS = {}
_log = logging.getLogger(__name__)


class WordlistType(type):
    def __new__(cls, name, bases, attrs):
        if bases:
            if "language_name" not in attrs:
                raise TypeError("Missing language_name for {0}".format(name))
            if "unique_prefix_length" not in attrs:
                raise TypeError("Missing 'unique_prefix_length' for {0}".format(name))
            if "word_list" not in attrs:
                raise TypeError("Missing 'word_list' for {0}".format(name))

            if "english_language_name" not in attrs:
                _log.warn(
                    "No 'english_language_name' for {0} using '{1}'".format(
                        name, language_name
                    )
                )
                attrs["english_language_name"] = attrs["language_name"]

            if len(attrs["word_list"]) != 1626:
                raise TypeError("Wrong word list length for {0}".format(name))

        new_cls = super(WordlistType, cls).__new__(cls, name, bases, attrs)

        if bases:
            WORDLISTS[new_cls.english_language_name] = new_cls

        return new_cls


class Wordlist(metaclass=WordlistType):
    n = 1626

    @classmethod
    def encode(cls, hex):
        """Convert hexadecimal string to mnemonic word representation with checksum."""
        out = []
        for i in range(len(hex) // 8):
            word = endian_swap(hex[8 * i : 8 * i + 8])
            x = int(word, 16)
            w1 = x % cls.n
            w2 = (x // cls.n + w1) % cls.n
            w3 = (x // cls.n // cls.n + w2) % cls.n
            out += [cls.word_list[w1], cls.word_list[w2], cls.word_list[w3]]
        checksum = cls.get_checksum(" ".join(out))
        out.append(checksum)
        return " ".join(out)

    @classmethod
    def decode(cls, phrase):
        """Calculate hexadecimal representation of the phrase."""
        phrase = phrase.split(" ")
        out = ""
        for i in range(len(phrase) // 3):
            word1, word2, word3 = phrase[3 * i : 3 * i + 3]
            w1 = cls.word_list.index(word1)
            w2 = cls.word_list.index(word2) % cls.n
            w3 = cls.word_list.index(word3) % cls.n
            x = w1 + cls.n * ((w2 - w1) % cls.n) + cls.n * cls.n * ((w3 - w2) % cls.n)
            out += endian_swap("%08x" % x)
        return out

    @classmethod
    def get_checksum(cls, phrase):
        """Given a mnemonic word string, return a string of the computed checksum.

        :rtype: str
        """
        phrase_split = phrase.split(" ")
        if len(phrase_split) < 12:
            raise ValueError("Invalid mnemonic phrase")
        if len(phrase_split) > 13:
            # Standard format
            phrase = phrase_split[:24]
        else:
            # MyMonero format
            phrase = phrase_split[:12]
        wstr = "".join(word[: cls.unique_prefix_length] for word in phrase)
        wstr = bytearray(wstr.encode("utf-8"))
        z = ((crc32(wstr) & 0xFFFFFFFF) ^ 0xFFFFFFFF) >> 0
        z2 = ((z ^ 0xFFFFFFFF) >> 0) % len(phrase)
        return phrase_split[z2]


def get_wordlist(name):
    try:
        return WORDLISTS[name]
    except KeyError:
        raise ValueError("No such word list")


def list_wordlists():
    return WORDLISTS.keys()


def endian_swap(word):
    """Given any string, swap bits and return the result.

    :rtype: str
    """
    return "".join([word[i : i + 2] for i in [6, 4, 2, 0]])



import binascii
import nacl.bindings

edwards_add = nacl.bindings.crypto_core_ed25519_add
inv = nacl.bindings.crypto_core_ed25519_scalar_invert
scalar_add = nacl.bindings.crypto_core_ed25519_scalar_add
scalarmult_B = nacl.bindings.crypto_scalarmult_ed25519_base_noclamp
scalarmult = nacl.bindings.crypto_scalarmult_ed25519_noclamp

# https://github.com/monero-project/monero/blob/9f814edbd78c70c70b814ca934c1ddef58768262/src/ringct/rctTypes.h#L615
H = binascii.unhexlify(
    "8b655970153799af2aeadc9ff1add0ea6c7251d54154cfa92c173a0dd39c1f94"
)


def scalarmult_H(v):
    return scalarmult(v, H)


def scalar_reduce(v):
    return nacl.bindings.crypto_core_ed25519_scalar_reduce(v + (64 - len(v)) * b"\0")


def public_from_secret_hex(hk):
    try:
        return binascii.hexlify(scalarmult_B(binascii.unhexlify(hk))).decode()
    except nacl.exceptions.RuntimeError:
        raise ValueError("Invalid secret key")



from .backends.jsonrpc import JSONRPCDaemon


class Daemon(object):
    """Monero daemon.

    Provides interface to a daemon instance.

    :param backend: a daemon backend
    :param \\**kwargs: arguments to initialize a :class:`JSONRPCDaemon <monero.backends.jsonrpc.JSONRPCDaemon>`
                        instance if no backend is given
    """

    def __init__(self, backend=None, **kwargs):
        if backend and len(kwargs):
            raise ValueError("backend already given, other arguments are extraneous")

        self._backend = backend if backend else JSONRPCDaemon(**kwargs)

    def info(self):
        """
        Returns basic information about the daemon.

        :rtype: dict
        """
        return self._backend.info()

    @property
    def net(self):
        return self._backend.net()

    def height(self):
        """
        Return daemon's chain height.

        :rtype: int
        """
        return self._backend.info()["height"]

    def send_transaction(self, tx, relay=True):
        """
        Sends a transaction generated by a :class:`Wallet <monero.wallet.Wallet>`.

        :param tx: :class:`Transaction <monero.transaction.Transaction>`
        :param relay: whether to relay the transaction to peers. If `False`, the daemon will have
                to mine the transaction itself in order to have it included in the blockchain.
        """
        return self._backend.send_transaction(tx.blob, relay=relay)

    def mempool(self):
        """
        Returns current mempool contents.

        :rtype: list of :class:`Transaction <monero.transaction.Transaction>`
        """
        return self._backend.mempool()

    def headers(self, start_height, end_height=None):
        """
        Returns block headers for given height range.
        If no :param end_height: is given, it's assumed to be equal to :param start_height:

        :rtype: list of dict
        """
        return self._backend.headers(start_height, end_height)

    def block(self, bhash=None, height=None):
        """
        Returns a block of specified height or hash.

        :param str bhash: block hash, or
        :param int height: block height

        :rtype: :class:`Block <monero.block.Block>`
        """
        if height is None and bhash is None:
            raise ValueError("Height or hash must be specified")
        return self._backend.block(bhash=bhash, height=height)

    def transactions(self, hashes):
        """
        Returns transactions matching given hashes. Accepts single hash or a sequence.

        :param hashes: str or list of str
        """
        if isinstance(hashes, str):
            hashes = [hashes]
        return self._backend.transactions(hashes)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Word list originally created by dabura667 and released under The MIT License (MIT)
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# Code surrounding the word list is Copyright (c) 2014-2018, The Monero Project
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are
# permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of
#    conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list
#    of conditions and the following disclaimer in the documentation and/or other
#    materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be
#    used to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# list of words from:
#     https://github.com/monero-project/monero/blob/master/src/mnemonics/esperanto.h
#     Most recent commit: 8ea3c4d54429976e9001b8cc4868b721e6747e3c


from .wordlist import Wordlist


class Esperanto(Wordlist):
    language_name = "Esperanto"
    english_language_name = "Esperanto"
    unique_prefix_length = 4
    word_list = [
        "abako",
        "abdiki",
        "abelo",
        "abituriento",
        "ablativo",
        "abnorma",
        "abonantoj",
        "abrikoto",
        "absoluta",
        "abunda",
        "acetono",
        "acida",
        "adapti",
        "adekvata",
        "adheri",
        "adicii",
        "adjektivo",
        "administri",
        "adolesko",
        "adreso",
        "adstringa",
        "adulto",
        "advokato",
        "adzo",
        "aeroplano",
        "aferulo",
        "afgana",
        "afiksi",
        "aflaba",
        "aforismo",
        "afranki",
        "aftozo",
        "afusto",
        "agavo",
        "agento",
        "agiti",
        "aglo",
        "agmaniero",
        "agnoski",
        "agordo",
        "agrabla",
        "agtipo",
        "agutio",
        "aikido",
        "ailanto",
        "aina",
        "ajatolo",
        "ajgenvaloro",
        "ajlobulbo",
        "ajnlitera",
        "ajuto",
        "ajzi",
        "akademio",
        "akcepti",
        "akeo",
        "akiri",
        "aklamado",
        "akmeo",
        "akno",
        "akompani",
        "akrobato",
        "akselo",
        "aktiva",
        "akurata",
        "akvofalo",
        "alarmo",
        "albumo",
        "alcedo",
        "aldoni",
        "aleo",
        "alfabeto",
        "algo",
        "alhasti",
        "aligatoro",
        "alkoholo",
        "almozo",
        "alnomo",
        "alojo",
        "alpinisto",
        "alrigardi",
        "alskribi",
        "alta",
        "alumeto",
        "alveni",
        "alzaca",
        "amaso",
        "ambasado",
        "amdeklaro",
        "amebo",
        "amfibio",
        "amhara",
        "amiko",
        "amkanto",
        "amletero",
        "amnestio",
        "amoranto",
        "amplekso",
        "amrakonto",
        "amsterdama",
        "amuzi",
        "ananaso",
        "androido",
        "anekdoto",
        "anfrakto",
        "angulo",
        "anheli",
        "animo",
        "anjono",
        "ankro",
        "anonci",
        "anpriskribo",
        "ansero",
        "antikva",
        "anuitato",
        "aorto",
        "aparta",
        "aperti",
        "apika",
        "aplikado",
        "apneo",
        "apogi",
        "aprobi",
        "apsido",
        "apterigo",
        "apudesto",
        "araneo",
        "arbo",
        "ardeco",
        "aresti",
        "argilo",
        "aristokrato",
        "arko",
        "arlekeno",
        "armi",
        "arniko",
        "aromo",
        "arpio",
        "arsenalo",
        "artisto",
        "aruba",
        "arvorto",
        "asaio",
        "asbesto",
        "ascendi",
        "asekuri",
        "asfalto",
        "asisti",
        "askalono",
        "asocio",
        "aspekti",
        "astro",
        "asulo",
        "atakonto",
        "atendi",
        "atingi",
        "atleto",
        "atmosfero",
        "atomo",
        "atropino",
        "atuto",
        "avataro",
        "aventuro",
        "aviadilo",
        "avokado",
        "azaleo",
        "azbuko",
        "azenino",
        "azilpetanto",
        "azoto",
        "azteka",
        "babili",
        "bacilo",
        "badmintono",
        "bagatelo",
        "bahama",
        "bajoneto",
        "baki",
        "balai",
        "bambuo",
        "bani",
        "baobabo",
        "bapti",
        "baro",
        "bastono",
        "batilo",
        "bavara",
        "bazalto",
        "beata",
        "bebofono",
        "bedo",
        "begonio",
        "behaviorismo",
        "bejlo",
        "bekero",
        "belarto",
        "bemolo",
        "benko",
        "bereto",
        "besto",
        "betulo",
        "bevelo",
        "bezoni",
        "biaso",
        "biblioteko",
        "biciklo",
        "bidaro",
        "bieno",
        "bifsteko",
        "bigamiulo",
        "bijekcio",
        "bikino",
        "bildo",
        "bimetalismo",
        "bindi",
        "biografio",
        "birdo",
        "biskvito",
        "bitlibro",
        "bivako",
        "bizara",
        "bjalistoka",
        "blanka",
        "bleki",
        "blinda",
        "blovi",
        "blua",
        "boato",
        "bobsledo",
        "bocvanano",
        "bodisatvo",
        "bofratino",
        "bogefratoj",
        "bohema",
        "boji",
        "bokalo",
        "boli",
        "bombono",
        "bona",
        "bopatrino",
        "bordo",
        "bosko",
        "botelo",
        "bovido",
        "brakpleno",
        "bretaro",
        "brikmuro",
        "broso",
        "brulema",
        "bubalo",
        "buctrapi",
        "budo",
        "bufedo",
        "bugio",
        "bujabeso",
        "buklo",
        "buldozo",
        "bumerango",
        "bunta",
        "burokrataro",
        "busbileto",
        "butero",
        "buzuko",
        "caro",
        "cebo",
        "ceceo",
        "cedro",
        "cefalo",
        "cejana",
        "cekumo",
        "celebri",
        "cemento",
        "cent",
        "cepo",
        "certa",
        "cetera",
        "cezio",
        "ciano",
        "cibeto",
        "cico",
        "cidro",
        "cifero",
        "cigaredo",
        "ciklo",
        "cilindro",
        "cimbalo",
        "cinamo",
        "cipreso",
        "cirkonstanco",
        "cisterno",
        "citrono",
        "ciumi",
        "civilizado",
        "colo",
        "congo",
        "cunamo",
        "cvana",
        "dabi",
        "daco",
        "dadaismo",
        "dafodilo",
        "dago",
        "daimio",
        "dajmono",
        "daktilo",
        "dalio",
        "damo",
        "danki",
        "darmo",
        "datumoj",
        "dazipo",
        "deadmoni",
        "debeto",
        "decidi",
        "dedukti",
        "deerigi",
        "defendi",
        "degeli",
        "dehaki",
        "deirpunkto",
        "deklaracio",
        "delikata",
        "demandi",
        "dento",
        "dependi",
        "derivi",
        "desegni",
        "detrui",
        "devi",
        "deziri",
        "dialogo",
        "dicentro",
        "didaktika",
        "dieto",
        "diferenci",
        "digesti",
        "diino",
        "dikfingro",
        "diligenta",
        "dimensio",
        "dinamo",
        "diodo",
        "diplomo",
        "direkte",
        "diskuti",
        "diurno",
        "diversa",
        "dizajno",
        "dobrogitaro",
        "docento",
        "dogano",
        "dojeno",
        "doktoro",
        "dolori",
        "domego",
        "donaci",
        "dopado",
        "dormi",
        "dosierujo",
        "dotita",
        "dozeno",
        "drato",
        "dresi",
        "drinki",
        "droni",
        "druido",
        "duaranga",
        "dubi",
        "ducent",
        "dudek",
        "duelo",
        "dufoje",
        "dugongo",
        "duhufa",
        "duilo",
        "dujare",
        "dukato",
        "duloka",
        "dumtempe",
        "dungi",
        "duobla",
        "dupiedulo",
        "dura",
        "dusenca",
        "dutaga",
        "duuma",
        "duvalvuloj",
        "duzo",
        "ebena",
        "eblecoj",
        "ebono",
        "ebria",
        "eburo",
        "ecaro",
        "ecigi",
        "ecoj",
        "edelvejso",
        "editoro",
        "edro",
        "eduki",
        "edzino",
        "efektiva",
        "efiki",
        "efloreski",
        "egala",
        "egeco",
        "egiptologo",
        "eglefino",
        "egoista",
        "egreto",
        "ejakuli",
        "ejlo",
        "ekarto",
        "ekbruligi",
        "ekceli",
        "ekde",
        "ekesti",
        "ekfirmao",
        "ekgliti",
        "ekhavi",
        "ekipi",
        "ekkapti",
        "eklezio",
        "ekmalsati",
        "ekonomio",
        "ekpluvi",
        "ekrano",
        "ekster",
        "ektiri",
        "ekumeno",
        "ekvilibro",
        "ekzemplo",
        "elasta",
        "elbalai",
        "elcento",
        "eldoni",
        "elektro",
        "elfari",
        "elgliti",
        "elhaki",
        "elipso",
        "elkovi",
        "ellasi",
        "elmeti",
        "elnutri",
        "elokventa",
        "elparoli",
        "elrevigi",
        "elstari",
        "elteni",
        "eluzita",
        "elvoki",
        "elzasa",
        "emajlo",
        "embaraso",
        "emerito",
        "emfazo",
        "eminenta",
        "emocio",
        "empiria",
        "emulsio",
        "enarkivigi",
        "enboteligi",
        "enciklopedio",
        "endorfino",
        "energio",
        "enfermi",
        "engluti",
        "enhavo",
        "enigmo",
        "enjekcio",
        "enketi",
        "enlanda",
        "enmeti",
        "enorma",
        "enplanti",
        "enradiki",
        "enspezo",
        "entrepreni",
        "enui",
        "envolvi",
        "enzimo",
        "eono",
        "eosto",
        "epitafo",
        "epoko",
        "epriskribebla",
        "epsilono",
        "erari",
        "erbio",
        "erco",
        "erekti",
        "ergonomia",
        "erikejo",
        "ermito",
        "erotika",
        "erpilo",
        "erupcio",
        "esameno",
        "escepti",
        "esenco",
        "eskapi",
        "esotera",
        "esperi",
        "estonto",
        "etapo",
        "etendi",
        "etfingro",
        "etikedo",
        "etlitero",
        "etmakleristo",
        "etnika",
        "etoso",
        "etradio",
        "etskala",
        "etullernejo",
        "evakui",
        "evento",
        "eviti",
        "evolui",
        "ezoko",
        "fabriko",
        "facila",
        "fadeno",
        "fagoto",
        "fajro",
        "fakto",
        "fali",
        "familio",
        "fanatiko",
        "farbo",
        "fasko",
        "fatala",
        "favora",
        "fazeolo",
        "febro",
        "federacio",
        "feino",
        "fekunda",
        "felo",
        "femuro",
        "fenestro",
        "fermi",
        "festi",
        "fetora",
        "fezo",
        "fiasko",
        "fibro",
        "fidela",
        "fiera",
        "fifama",
        "figuro",
        "fiherbo",
        "fiinsekto",
        "fiksa",
        "filmo",
        "fimensa",
        "finalo",
        "fiolo",
        "fiparoli",
        "firmao",
        "fisko",
        "fitingo",
        "fiuzanto",
        "fivorto",
        "fiziko",
        "fjordo",
        "flago",
        "flegi",
        "flirti",
        "floro",
        "flugi",
        "fobio",
        "foceno",
        "foirejo",
        "fojfoje",
        "fokuso",
        "folio",
        "fomenti",
        "fonto",
        "formulo",
        "fosforo",
        "fotografi",
        "fratino",
        "fremda",
        "friti",
        "frosto",
        "frua",
        "ftizo",
        "fuelo",
        "fugo",
        "fuksia",
        "fulmilo",
        "fumanto",
        "fundamento",
        "fuorto",
        "furioza",
        "fusilo",
        "futbalo",
        "fuzio",
        "gabardino",
        "gado",
        "gaela",
        "gafo",
        "gagato",
        "gaja",
        "gaki",
        "galanta",
        "gamao",
        "ganto",
        "gapulo",
        "gardi",
        "gasto",
        "gavio",
        "gazeto",
        "geamantoj",
        "gebani",
        "geedzeco",
        "gefratoj",
        "geheno",
        "gejsero",
        "geko",
        "gelateno",
        "gemisto",
        "geniulo",
        "geografio",
        "gepardo",
        "geranio",
        "gestolingvo",
        "geto",
        "geumo",
        "gibono",
        "giganta",
        "gildo",
        "gimnastiko",
        "ginekologo",
        "gipsi",
        "girlando",
        "gistfungo",
        "gitaro",
        "glazuro",
        "glebo",
        "gliti",
        "globo",
        "gluti",
        "gnafalio",
        "gnejso",
        "gnomo",
        "gnuo",
        "gobio",
        "godetio",
        "goeleto",
        "gojo",
        "golfludejo",
        "gombo",
        "gondolo",
        "gorilo",
        "gospelo",
        "gotika",
        "granda",
        "greno",
        "griza",
        "groto",
        "grupo",
        "guano",
        "gubernatoro",
        "gudrotuko",
        "gufo",
        "gujavo",
        "guldeno",
        "gumi",
        "gupio",
        "guruo",
        "gusto",
        "guto",
        "guvernistino",
        "gvardio",
        "gverilo",
        "gvidanto",
        "habitato",
        "hadito",
        "hafnio",
        "hagiografio",
        "haitiano",
        "hajlo",
        "hakbloko",
        "halti",
        "hamstro",
        "hangaro",
        "hapalo",
        "haro",
        "hasta",
        "hati",
        "havebla",
        "hazardo",
        "hebrea",
        "hedero",
        "hegemonio",
        "hejmo",
        "hektaro",
        "helpi",
        "hemisfero",
        "heni",
        "hepato",
        "herbo",
        "hesa",
        "heterogena",
        "heziti",
        "hiacinto",
        "hibrida",
        "hidrogeno",
        "hieroglifo",
        "higieno",
        "hihii",
        "hilumo",
        "himno",
        "hindino",
        "hiperteksto",
        "hirundo",
        "historio",
        "hobio",
        "hojli",
        "hokeo",
        "hologramo",
        "homido",
        "honesta",
        "hopi",
        "horizonto",
        "hospitalo",
        "hotelo",
        "huadi",
        "hubo",
        "hufumo",
        "hugenoto",
        "hukero",
        "huligano",
        "humana",
        "hundo",
        "huoj",
        "hupilo",
        "hurai",
        "husaro",
        "hutuo",
        "huzo",
        "iafoje",
        "iagrade",
        "iamaniere",
        "iarelate",
        "iaspeca",
        "ibekso",
        "ibiso",
        "idaro",
        "ideala",
        "idiomo",
        "idolo",
        "iele",
        "igluo",
        "ignori",
        "iguamo",
        "igvano",
        "ikono",
        "iksodo",
        "ikto",
        "iliaflanke",
        "ilkomputilo",
        "ilobreto",
        "ilremedo",
        "ilumini",
        "imagi",
        "imitado",
        "imperio",
        "imuna",
        "incidento",
        "industrio",
        "inerta",
        "infano",
        "ingenra",
        "inhali",
        "iniciati",
        "injekti",
        "inklino",
        "inokuli",
        "insekto",
        "inteligenta",
        "inundi",
        "inviti",
        "ioma",
        "ionosfero",
        "iperito",
        "ipomeo",
        "irana",
        "irejo",
        "irigacio",
        "ironio",
        "isato",
        "islamo",
        "istempo",
        "itinero",
        "itrio",
        "iuloke",
        "iumaniere",
        "iutempe",
        "izolita",
        "jado",
        "jaguaro",
        "jakto",
        "jama",
        "januaro",
        "japano",
        "jarringo",
        "jazo",
        "jenoj",
        "jesulo",
        "jetavio",
        "jezuito",
        "jodli",
        "joviala",
        "juano",
        "jubileo",
        "judismo",
        "jufto",
        "juki",
        "julio",
        "juneca",
        "jupo",
        "juristo",
        "juste",
        "juvelo",
        "kabineto",
        "kadrato",
        "kafo",
        "kahelo",
        "kajako",
        "kakao",
        "kalkuli",
        "kampo",
        "kanti",
        "kapitalo",
        "karaktero",
        "kaserolo",
        "katapulto",
        "kaverna",
        "kazino",
        "kebabo",
        "kefiro",
        "keglo",
        "kejlo",
        "kekso",
        "kelka",
        "kemio",
        "kerno",
        "kesto",
        "kiamaniere",
        "kibuco",
        "kidnapi",
        "kielo",
        "kikero",
        "kilogramo",
        "kimono",
        "kinejo",
        "kiosko",
        "kirurgo",
        "kisi",
        "kitelo",
        "kivio",
        "klavaro",
        "klerulo",
        "klini",
        "klopodi",
        "klubo",
        "knabo",
        "knedi",
        "koalo",
        "kobalto",
        "kodigi",
        "kofro",
        "kohera",
        "koincidi",
        "kojoto",
        "kokoso",
        "koloro",
        "komenci",
        "kontrakto",
        "kopio",
        "korekte",
        "kosti",
        "kotono",
        "kovri",
        "krajono",
        "kredi",
        "krii",
        "krom",
        "kruco",
        "ksantino",
        "ksenono",
        "ksilofono",
        "ksosa",
        "kubuto",
        "kudri",
        "kuglo",
        "kuiri",
        "kuko",
        "kulero",
        "kumuluso",
        "kuneco",
        "kupro",
        "kuri",
        "kuseno",
        "kutimo",
        "kuvo",
        "kuzino",
        "kvalito",
        "kverko",
        "kvin",
        "kvoto",
        "labori",
        "laculo",
        "ladbotelo",
        "lafo",
        "laguno",
        "laikino",
        "laktobovino",
        "lampolumo",
        "landkarto",
        "laosa",
        "lapono",
        "larmoguto",
        "lastjare",
        "latitudo",
        "lavejo",
        "lazanjo",
        "leciono",
        "ledosako",
        "leganto",
        "lekcio",
        "lemura",
        "lentuga",
        "leopardo",
        "leporo",
        "lerni",
        "lesivo",
        "letero",
        "levilo",
        "lezi",
        "liano",
        "libera",
        "liceo",
        "lieno",
        "lifto",
        "ligilo",
        "likvoro",
        "lila",
        "limono",
        "lingvo",
        "lipo",
        "lirika",
        "listo",
        "literatura",
        "liveri",
        "lobio",
        "logika",
        "lojala",
        "lokalo",
        "longa",
        "lordo",
        "lotado",
        "loza",
        "luanto",
        "lubriki",
        "lucida",
        "ludema",
        "luigi",
        "lukso",
        "luli",
        "lumbilda",
        "lunde",
        "lupago",
        "lustro",
        "lutilo",
        "luzerno",
        "maato",
        "maceri",
        "madono",
        "mafiano",
        "magazeno",
        "mahometano",
        "maizo",
        "majstro",
        "maketo",
        "malgranda",
        "mamo",
        "mandareno",
        "maorio",
        "mapigi",
        "marini",
        "masko",
        "mateno",
        "mazuto",
        "meandro",
        "meblo",
        "mecenato",
        "medialo",
        "mefito",
        "megafono",
        "mejlo",
        "mekanika",
        "melodia",
        "membro",
        "mendi",
        "mergi",
        "mespilo",
        "metoda",
        "mevo",
        "mezuri",
        "miaflanke",
        "micelio",
        "mielo",
        "migdalo",
        "mikrofilmo",
        "militi",
        "mimiko",
        "mineralo",
        "miopa",
        "miri",
        "mistera",
        "mitralo",
        "mizeri",
        "mjelo",
        "mnemoniko",
        "mobilizi",
        "mocio",
        "moderna",
        "mohajro",
        "mokadi",
        "molaro",
        "momento",
        "monero",
        "mopso",
        "mordi",
        "moskito",
        "motoro",
        "movimento",
        "mozaiko",
        "mueli",
        "mukozo",
        "muldi",
        "mumio",
        "munti",
        "muro",
        "muskolo",
        "mutacio",
        "muzikisto",
        "nabo",
        "nacio",
        "nadlo",
        "nafto",
        "naiva",
        "najbaro",
        "nanometro",
        "napo",
        "narciso",
        "naski",
        "naturo",
        "navigi",
        "naztruo",
        "neatendite",
        "nebulo",
        "necesa",
        "nedankinde",
        "neebla",
        "nefari",
        "negoco",
        "nehavi",
        "neimagebla",
        "nektaro",
        "nelonga",
        "nematura",
        "nenia",
        "neordinara",
        "nepra",
        "nervuro",
        "nesto",
        "nete",
        "neulo",
        "nevino",
        "nifo",
        "nigra",
        "nihilisto",
        "nikotino",
        "nilono",
        "nimfeo",
        "nitrogeno",
        "nivelo",
        "nobla",
        "nocio",
        "nodozo",
        "nokto",
        "nomkarto",
        "norda",
        "nostalgio",
        "notbloko",
        "novico",
        "nuanco",
        "nuboza",
        "nuda",
        "nugato",
        "nuklea",
        "nuligi",
        "numero",
        "nuntempe",
        "nupto",
        "nura",
        "nutri",
        "oazo",
        "obei",
        "objekto",
        "oblikva",
        "obolo",
        "observi",
        "obtuza",
        "obuso",
        "oceano",
        "odekolono",
        "odori",
        "oferti",
        "oficiala",
        "ofsajdo",
        "ofte",
        "ogivo",
        "ogro",
        "ojstredoj",
        "okaze",
        "okcidenta",
        "okro",
        "oksido",
        "oktobro",
        "okulo",
        "oldulo",
        "oleo",
        "olivo",
        "omaro",
        "ombro",
        "omego",
        "omikrono",
        "omleto",
        "omnibuso",
        "onagro",
        "ondo",
        "oneco",
        "onidire",
        "onklino",
        "onlajna",
        "onomatopeo",
        "ontologio",
        "opaka",
        "operacii",
        "opinii",
        "oportuna",
        "opresi",
        "optimisto",
        "oratoro",
        "orbito",
        "ordinara",
        "orelo",
        "orfino",
        "organizi",
        "orienta",
        "orkestro",
        "orlo",
        "orminejo",
        "ornami",
        "ortangulo",
        "orumi",
        "oscedi",
        "osmozo",
        "ostocerbo",
        "ovalo",
        "ovingo",
        "ovoblanko",
        "ovri",
        "ovulado",
        "ozono",
        "pacama",
        "padeli",
        "pafilo",
        "pagigi",
        "pajlo",
        "paketo",
        "palaco",
        "pampelmo",
        "pantalono",
        "papero",
        "paroli",
        "pasejo",
        "patro",
        "pavimo",
        "peco",
        "pedalo",
        "peklita",
        "pelikano",
        "pensiono",
        "peplomo",
        "pesilo",
        "petanto",
        "pezoforto",
        "piano",
        "picejo",
        "piede",
        "pigmento",
        "pikema",
        "pilkoludo",
        "pimento",
        "pinglo",
        "pioniro",
        "pipromento",
        "pirato",
        "pistolo",
        "pitoreska",
        "piulo",
        "pivoti",
        "pizango",
        "planko",
        "plektita",
        "plibonigi",
        "ploradi",
        "plurlingva",
        "pobo",
        "podio",
        "poeto",
        "pogranda",
        "pohora",
        "pokalo",
        "politekniko",
        "pomarbo",
        "ponevosto",
        "populara",
        "porcelana",
        "postkompreno",
        "poteto",
        "poviga",
        "pozitiva",
        "prapatroj",
        "precize",
        "pridemandi",
        "probable",
        "pruntanto",
        "psalmo",
        "psikologio",
        "psoriazo",
        "pterido",
        "publiko",
        "pudro",
        "pufo",
        "pugnobato",
        "pulovero",
        "pumpi",
        "punkto",
        "pupo",
        "pureo",
        "puso",
        "putrema",
        "puzlo",
        "rabate",
        "racionala",
        "radiko",
        "rafinado",
        "raguo",
        "rajto",
        "rakonti",
        "ralio",
        "rampi",
        "rando",
        "rapida",
        "rastruma",
        "ratifiki",
        "raviolo",
        "razeno",
        "reakcio",
        "rebildo",
        "recepto",
        "redakti",
        "reenigi",
        "reformi",
        "regiono",
        "rehavi",
        "reinspekti",
        "rejesi",
        "reklamo",
        "relativa",
        "rememori",
        "renkonti",
        "reorganizado",
        "reprezenti",
        "respondi",
        "retumilo",
        "reuzebla",
        "revidi",
        "rezulti",
        "rialo",
        "ribeli",
        "ricevi",
        "ridiga",
        "rifuginto",
        "rigardi",
        "rikolti",
        "rilati",
        "rimarki",
        "rinocero",
        "ripozi",
        "riski",
        "ritmo",
        "rivero",
        "rizokampo",
        "roboto",
        "rododendro",
        "rojo",
        "rokmuziko",
        "rolvorto",
        "romantika",
        "ronroni",
        "rosino",
        "rotondo",
        "rovero",
        "rozeto",
        "rubando",
        "rudimenta",
        "rufa",
        "rugbeo",
        "ruino",
        "ruleto",
        "rumoro",
        "runo",
        "rupio",
        "rura",
        "rustimuna",
        "ruzulo",
        "sabato",
        "sadismo",
        "safario",
        "sagaca",
        "sakfluto",
        "salti",
        "samtage",
        "sandalo",
        "sapejo",
        "sarongo",
        "satelito",
        "savano",
        "sbiro",
        "sciado",
        "seanco",
        "sebo",
        "sedativo",
        "segligno",
        "sekretario",
        "selektiva",
        "semajno",
        "senpeza",
        "separeo",
        "servilo",
        "sesangulo",
        "setli",
        "seurigi",
        "severa",
        "sezono",
        "sfagno",
        "sfero",
        "sfinkso",
        "siatempe",
        "siblado",
        "sidejo",
        "siesto",
        "sifono",
        "signalo",
        "siklo",
        "silenti",
        "simpla",
        "sinjoro",
        "siropo",
        "sistemo",
        "situacio",
        "siverto",
        "sizifa",
        "skatolo",
        "skemo",
        "skianto",
        "sklavo",
        "skorpio",
        "skribisto",
        "skulpti",
        "skvamo",
        "slango",
        "sledeto",
        "sliparo",
        "smeraldo",
        "smirgi",
        "smokingo",
        "smuto",
        "snoba",
        "snufegi",
        "sobra",
        "sociano",
        "sodakvo",
        "sofo",
        "soifi",
        "sojlo",
        "soklo",
        "soldato",
        "somero",
        "sonilo",
        "sopiri",
        "sorto",
        "soulo",
        "soveto",
        "sparkado",
        "speciala",
        "spiri",
        "splito",
        "sporto",
        "sprita",
        "spuro",
        "stabila",
        "stelfiguro",
        "stimulo",
        "stomako",
        "strato",
        "studanto",
        "subgrupo",
        "suden",
        "suferanta",
        "sugesti",
        "suito",
        "sukero",
        "sulko",
        "sume",
        "sunlumo",
        "super",
        "surskribeto",
        "suspekti",
        "suturo",
        "svati",
        "svenfali",
        "svingi",
        "svopo",
        "tabako",
        "taglumo",
        "tajloro",
        "taksimetro",
        "talento",
        "tamen",
        "tanko",
        "taoismo",
        "tapioko",
        "tarifo",
        "tasko",
        "tatui",
        "taverno",
        "teatro",
        "tedlaboro",
        "tegmento",
        "tehoro",
        "teknika",
        "telefono",
        "tempo",
        "tenisejo",
        "teorie",
        "teraso",
        "testudo",
        "tetablo",
        "teujo",
        "tezo",
        "tialo",
        "tibio",
        "tielnomata",
        "tifono",
        "tigro",
        "tikli",
        "timida",
        "tinkturo",
        "tiom",
        "tiparo",
        "tirkesto",
        "titolo",
        "tiutempe",
        "tizano",
        "tobogano",
        "tofeo",
        "togo",
        "toksa",
        "tolerema",
        "tombolo",
        "tondri",
        "topografio",
        "tordeti",
        "tosti",
        "totalo",
        "traduko",
        "tredi",
        "triangulo",
        "tropika",
        "trumpeto",
        "tualeto",
        "tubisto",
        "tufgrebo",
        "tuja",
        "tukano",
        "tulipo",
        "tumulto",
        "tunelo",
        "turisto",
        "tusi",
        "tutmonda",
        "tvisto",
        "udono",
        "uesto",
        "ukazo",
        "ukelelo",
        "ulcero",
        "ulmo",
        "ultimato",
        "ululi",
        "umbiliko",
        "unco",
        "ungego",
        "uniformo",
        "unkti",
        "unukolora",
        "uragano",
        "urbano",
        "uretro",
        "urino",
        "ursido",
        "uskleco",
        "usonigi",
        "utero",
        "utila",
        "utopia",
        "uverturo",
        "uzadi",
        "uzeblo",
        "uzino",
        "uzkutimo",
        "uzofini",
        "uzurpi",
        "uzvaloro",
        "vadejo",
        "vafleto",
        "vagono",
        "vahabismo",
        "vajco",
        "vakcino",
        "valoro",
        "vampiro",
        "vangharoj",
        "vaporo",
        "varma",
        "vasta",
        "vato",
        "vazaro",
        "veaspekta",
        "vedismo",
        "vegetalo",
        "vehiklo",
        "vejno",
        "vekita",
        "velstango",
        "vemieno",
        "vendi",
        "vepro",
        "verando",
        "vespero",
        "veturi",
        "veziko",
        "viando",
        "vibri",
        "vico",
        "videbla",
        "vifio",
        "vigla",
        "viktimo",
        "vila",
        "vimeno",
        "vintro",
        "violo",
        "vippuno",
        "virtuala",
        "viskoza",
        "vitro",
        "viveca",
        "viziti",
        "vobli",
        "vodko",
        "vojeto",
        "vokegi",
        "volbo",
        "vomema",
        "vono",
        "vortaro",
        "vosto",
        "voti",
        "vrako",
        "vringi",
        "vualo",
        "vulkano",
        "vundo",
        "vuvuzelo",
        "zamenhofa",
        "zapi",
        "zebro",
        "zefiro",
        "zeloto",
        "zenismo",
        "zeolito",
        "zepelino",
        "zeto",
        "zigzagi",
        "zinko",
        "zipo",
        "zirkonio",
        "zodiako",
        "zoeto",
        "zombio",
        "zono",
        "zoologio",
        "zorgi",
        "zukino",
        "zumilo",
    ]



import binascii
from datetime import datetime
from decimal import Decimal
import ipaddress
import json
import logging
import requests

from ... import address
from ... import exceptions
from ...block import Block
from ...const import NET_MAIN, NET_TEST, NET_STAGE
from ...numbers import from_atomic, to_atomic
from ...transaction import Transaction
from .exceptions import RPCError, MethodNotFound, Unauthorized


_log = logging.getLogger(__name__)


RESTRICTED_MAX_TRANSACTIONS = 100


class JSONRPCDaemon(object):
    """
    JSON RPC backend for Monero daemon.
    The offical documentation for the Monero Daemon RPC Protocol can be found at the link below.
    https://www.getmonero.org/resources/developer-guides/daemon-rpc.html#on_get_block_hash
    Much of the in-code documentation of this class is derived from this document.

    :param protocol: `http` or `https`
    :param host: host name or IP
    :param port: port number
    :param path: path for JSON RPC requests (should not be changed)
    :param timeout: request timeout
    :param verify_ssl_certs: verify SSL certificates when connecting
    :param proxy_url: a proxy to use
    :param prune_transactions: whether to prune transaction data. Saves bandwidth but you may want
                            to enable it when you need to retrieve transaction binary blobs.
    """

    _METHOD_NOT_FOUND_CODE = -32601
    _KNOWN_LOG_CATEGORIES = [
        "*",
        "default",
        "net",
        "net.http",
        "net.p2p",
        "logging",
        "net.throttle",
        "blockchain.db",
        "blockchain.db.lmdb",
        "bcutil",
        "checkpoints",
        "net.dns",
        "net.dl",
        "i18n",
        "perf",
        "stacktrace",
        "updates",
        "account",
        "cn",
        "difficulty",
        "hardfork",
        "miner",
        "blockchain",
        "txpool",
        "cn.block_queue",
        "net.cn",
        "daemon",
        "debugtools.deserialize",
        "debugtools.objectsizes",
        "device.ledger",
        "wallet.gen_multisig",
        "multisig",
        "bulletproofs",
        "ringct",
        "daemon.rpc",
        "wallet.simplewallet",
        "WalletAPI",
        "wallet.ringdb",
        "wallet.wallet2",
        "wallet.rpc" "tests.core",
    ]
    _KNOWN_LOG_LEVELS = ["FATAL", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE"]

    _net = None
    _restricted = None

    def __init__(
        self,
        protocol="http",
        host="127.0.0.1",
        port=18081,
        path="/json_rpc",
        user="",
        password="",
        timeout=30,
        verify_ssl_certs=True,
        proxy_url=None,
        prune_transactions=True,
    ):
        self.url = "{protocol}://{host}:{port}".format(
            protocol=protocol, host=host, port=port
        )
        _log.debug("JSONRPC daemon backend URL: {url}".format(url=self.url))
        self.auth = requests.auth.HTTPDigestAuth(user, password)
        self.session = requests.Session()
        self.timeout = timeout
        self.verify_ssl_certs = verify_ssl_certs
        self.proxies = {protocol: proxy_url}
        self.prune_transactions = prune_transactions

    def info(self):
        info = self.raw_jsonrpc_request("get_info")
        self._set_net(info)
        return info

    def net(self):
        if self._net:
            return self._net
        self.info()
        return self._net

    def restricted(self):
        if self._restricted is None:
            self._set_restricted()

        return self._restricted

    def send_transaction(self, blob, relay=True):
        res = self.raw_request(
            "/sendrawtransaction",
            {
                "tx_as_hex": binascii.hexlify(blob).decode(),
                "do_not_relay": not relay,
            },
        )
        if res["status"] == "OK":
            return res
        raise exceptions.TransactionBroadcastError(
            "{status}: {reason}".format(**res), details=res
        )

    def mempool(self):
        res = self.raw_request("/get_transaction_pool", {})
        txs = []
        for tx in res.get("transactions", []):
            txs.append(
                Transaction(
                    hash=tx["id_hash"],
                    fee=from_atomic(tx["fee"]),
                    timestamp=datetime.fromtimestamp(tx["receive_time"]),
                    blob=binascii.unhexlify(tx["tx_blob"]),
                    json=json.loads(tx["tx_json"]),
                    confirmations=0,
                )
            )
        return txs

    def headers(self, start_height, end_height=None):
        end_height = end_height or start_height
        res = self.raw_jsonrpc_request(
            "get_block_headers_range",
            {"start_height": start_height, "end_height": end_height},
        )
        if res["status"] == "OK":
            return res["headers"]
        raise exceptions.BackendException(res["status"])

    def block(self, bhash=None, height=None):
        data = {}
        if bhash:
            data["hash"] = bhash
        if height:
            data["height"] = height
        res = self.raw_jsonrpc_request("get_block", data)
        if res["status"] == "OK":
            bhdr = res["block_header"]
            sub_json = json.loads(res["json"])
            data = {
                "blob": res["blob"],
                "hash": bhdr["hash"],
                "height": bhdr["height"],
                "timestamp": datetime.fromtimestamp(bhdr["timestamp"]),
                "version": (bhdr["major_version"], bhdr["minor_version"]),
                "difficulty": bhdr["difficulty"],
                "nonce": bhdr["nonce"],
                "orphan": bhdr["orphan_status"],
                "prev_hash": bhdr["prev_hash"],
                "reward": from_atomic(bhdr["reward"]),
                "transactions": self.transactions(
                    [bhdr["miner_tx_hash"]] + sub_json["tx_hashes"]
                ),
            }
            return Block(**data)
        raise exceptions.BackendException(res["status"])

    def transactions(self, hashes):
        """
        Returns a list of transactions for given hashes. Automatically chunks the request
        into amounts acceptable by a restricted RPC server.
        """

        hashes = self._validate_hashlist(hashes)
        result = []
        while len(hashes):
            result.extend(
                self._do_get_transactions(
                    hashes[:RESTRICTED_MAX_TRANSACTIONS], prune=self.prune_transactions
                )
            )
            hashes = hashes[RESTRICTED_MAX_TRANSACTIONS:]
        return result

    def raw_request(self, path, data=None):
        hdr = {"Content-Type": "application/json"}
        _log.debug(
            "Request: {path}\nData: {data}".format(
                path=path, data=json.dumps(data, indent=2, sort_keys=True)
            )
        )
        rsp = self.session.post(
            self.url + path,
            headers=hdr,
            data=json.dumps(data) if data else None,
            auth=self.auth,
            timeout=self.timeout,
            verify=self.verify_ssl_certs,
            proxies=self.proxies,
        )
        if rsp.status_code != 200:
            raise RPCError(
                "Invalid HTTP status {code} for path {path}.".format(
                    code=rsp.status_code, path=path
                )
            )
        result = rsp.json()
        _ppresult = json.dumps(result, indent=2, sort_keys=True)
        _log.debug("Result:\n{result}".format(result=_ppresult))
        return result

    def raw_jsonrpc_request(self, method, params=None):
        hdr = {"Content-Type": "application/json"}
        data = {"jsonrpc": "2.0", "id": 0, "method": method, "params": params or {}}
        _log.debug(
            "Method: {method}\nParams:\n{params}".format(
                method=method, params=json.dumps(params, indent=2, sort_keys=True)
            )
        )
        rsp = self.session.post(
            self.url + "/json_rpc",
            headers=hdr,
            data=json.dumps(data),
            auth=self.auth,
            timeout=self.timeout,
            verify=self.verify_ssl_certs,
            proxies=self.proxies,
        )

        if rsp.status_code == 401:
            raise Unauthorized("401 Unauthorized. Invalid RPC user name or password.")
        elif rsp.status_code != 200:
            raise RPCError(
                "Invalid HTTP status {code} for method {method}.".format(
                    code=rsp.status_code, method=method
                )
            )

        try:
            result = rsp.json()
        except ValueError as e:
            _log.error(
                "Could not parse JSON response from '{url}' during method '{method}'. Response:\n{resp}".format(
                    url=self.url, method=method, resp=rsp.text
                )
            )
            raise RPCError(
                "Daemon returned an unreadable JSON response. It may contain unparseable binary characters."
            )

        _ppresult = json.dumps(result, indent=2, sort_keys=True)
        _log.debug("Result:\n{result}".format(result=_ppresult))

        if "error" in result:
            err = result["error"]
            code = err["code"]
            msg = err["message"]

            if code == self._METHOD_NOT_FOUND_CODE:
                raise MethodNotFound('Daemon method "{}" not found'.format(method))

            _log.error("JSON RPC error:\n{result}".format(result=_ppresult))
            raise RPCError(
                "Method '{method}' failed with RPC Error of unknown code {code}, "
                "message: {msg}".format(
                    method=method, data=data, result=result, code=code, msg=msg
                )
            )
        elif "status" in result["result"] and result["result"]["status"] == "BUSY":
            raise exceptions.DaemonIsBusy(
                "In JSONRPC method '{method}': daemon at {url} is in BUSY mode. RPC command results will be unpredictable "
                "until daemon is fully synced.".format(method=method, url=self.url)
            )

        return result["result"]

    # JSON RPC Methods (https://www.getmonero.org/resources/developer-guides/daemon-rpc.html#json-rpc-methods)

    def get_block_count(self):
        """
        Look up how many blocks are in the longest chain known to the node.

        Output:
        {
        "count": unsigned int; Number of blocks in longest chain seen by the node.
        "status": str; General RPC error code. "OK" means everything looks good.
        }
        """

        return self.raw_jsonrpc_request("get_block_count")

    def on_get_block_hash(self, height):
        """
        Look up a block's hash by its height.

        :param int height: height of the block

        Output: str; block hash
        """

        height = int(height)

        if height < 0:
            raise ValueError("height < 0")

        return self.raw_jsonrpc_request("on_get_block_hash", params=[height])

    def get_block_template(self, wallet_address, reserve_size):
        """
        Get a block template on which mining a new block.

        :param str wallet_address: Address of wallet to receive coinbase transactions for successful block.
        :param int reserve_size: number of bytes to make reserved space, maximum is 127

        Output:
        {
        "blocktemplate_blob": str; Blob on which to try to mine a new block.
        "blockhashing_blob": str; Blob on which to try to find a valid nonce.
        "difficulty": unsigned int; Bottom 64 bits of difficulty of next block.
        "difficuly_top64": unsigned int; Top 64 bits of difficulty of next block.
        "wide_difficulty": str; hex str of difficulty of next block.
        "expected_reward": unsigned int; Coinbase reward expected to be received if block is successfully mined.
        "height": unsigned int; Height on which to mine.
        "prev_hash": str; Hash of the most recent block on which to mine the next block.
        "reserved_offset": unsigned int; Offset in bytes to reserved space in block blob.
        "seed_hash": str;
        "seed_height": unsigned int;
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode
        }
        """

        try:
            address.address(wallet_address)
        except ValueError:
            raise ValueError("wallet_address is not in a recognized address form")

        if reserve_size not in range(128):
            raise ValueError("reserve_size {} is out of bounds".format(reserve_size))

        return self.raw_jsonrpc_request(
            "get_block_template",
            params={"wallet_address": wallet_address, "reserve_size": reserve_size},
        )

    def submit_block(self, blobs):
        """
        Submit a mined block to the network.

        :param list blobs: list of block hex strs or raw bytes objects

        Output:
        {
        "status": str; Block submit status
        }
        """

        if isinstance(blobs, (bytes, str)):
            blobs = [blobs]

        blobs = [
            binascii.hexlify(b).decode() if isinstance(b, bytes) else b for b in blobs
        ]

        return self.raw_jsonrpc_request("submit_block", params=blobs)

    def get_last_block_header(self):
        """
        Block header information for the most recent block is easily retrieved with this method.

        Output:
        {
        "block_header": {
            "block_size": unsigned int; The block size in bytes.
            "depth": unsigned int; The number of blocks succeeding this block on the blockchain. A larger number means an older block.
            "difficulty" unsigned int; The strength of the Monero network based on mining power.
            "hash": str; The hash of this block.
            "height": unsigned int; The number of blocks preceding this block on the blockchain.
            "major_version": unsigned int; The major version of the monero protocol at this block height.
            "minor_version": unsigned int; The minor version of the monero protocol at this block height.
            "nonce": unsigned int; a cryptographic random one-time number used in mining a Monero block.
            "num_txes": unsigned int; Number of transactions in the block, not counting the coinbase tx.
            "orphan_status": bool; Usually false. If true, this block is not part of the longest chain.
            "prev_hash": str; The hash of the block immediately preceding this block in the chain.
            "reward": unsigned int; The amount of new atomic units generated in this block and rewarded to the miner. Note: 1 XMR = 1e12 atomic units.
            "timestamp": unsigned int; The unix time at which the block was recorded into the blockchain.
        }
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode
        }
        """

        return self.raw_jsonrpc_request("get_last_block_header")

    def get_block_header_by_hash(self, blk_hash):
        """
        Block header information can be retrieved using a block's hash

        :param str blk_hash: block hash

        Output:
        {
        "block_header": {
            "block_size": unsigned int; The block size in bytes.
            "depth": unsigned int; The number of blocks succeeding this block on the blockchain. A larger number means an older block.
            "difficulty" unsigned int; The strength of the Monero network based on mining power.
            "hash": str; The hash of this block.
            "height": unsigned int; The number of blocks preceding this block on the blockchain.
            "major_version": unsigned int; The major version of the monero protocol at this block height.
            "minor_version": unsigned int; The minor version of the monero protocol at this block height.
            "nonce": unsigned int; a cryptographic random one-time number used in mining a Monero block.
            "num_txes": unsigned int; Number of transactions in the block, not counting the coinbase tx.
            "orphan_status": bool; Usually false. If true, this block is not part of the longest chain.
            "prev_hash": str; The hash of the block immediately preceding this block in the chain.
            "reward": unsigned int; The amount of new atomic units generated in this block and rewarded to the miner. Note: 1 XMR = 1e12 atomic units.
            "timestamp": unsigned int; The unix time at which the block was recorded into the blockchain.
        }
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode
        }
        """

        if not self._is_valid_256_hex(blk_hash):
            raise ValueError('blk_hash "{}" is not a valid hash'.format(blk_hash))

        return self.raw_jsonrpc_request(
            "get_block_header_by_hash", params={"hash": blk_hash}
        )

    def get_block_header_by_height(self, height):
        """
        Block header information can be retrieved using a block's height

        :param int height: block height

        Output:
        {
        "block_header": {
            "block_size": unsigned int; The block size in bytes.
            "depth": unsigned int; The number of blocks succeeding this block on the blockchain. A larger number means an older block.
            "difficulty" unsigned int; The strength of the Monero network based on mining power.
            "hash": str; The hash of this block.
            "height": unsigned int; The number of blocks preceding this block on the blockchain.
            "major_version": unsigned int; The major version of the monero protocol at this block height.
            "minor_version": unsigned int; The minor version of the monero protocol at this block height.
            "nonce": unsigned int; a cryptographic random one-time number used in mining a Monero block.
            "num_txes": unsigned int; Number of transactions in the block, not counting the coinbase tx.
            "orphan_status": bool; Usually false. If true, this block is not part of the longest chain.
            "prev_hash": str; The hash of the block immediately preceding this block in the chain.
            "reward": unsigned int; The amount of new atomic units generated in this block and rewarded to the miner. Note: 1 XMR = 1e12 atomic units.
            "timestamp": unsigned int; The unix time at which the block was recorded into the blockchain.
        }
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode
        }
        """

        if height < 0:
            raise ValueError("height < 0")

        return self.raw_jsonrpc_request(
            "get_block_header_by_height", params={"height": height}
        )

    def get_block_headers_range(self, start_height, end_height):
        """
        Similar to get_block_header_by_height, but for a range of blocks. The range is inclusive,
        so the number of blocks returned will be equal to end_height - start_height + 1

        :param int start_height: starting height
        :param int end_height: ending height

        Output:
        {
        "headers": list; block_header structures (get_last_block_header) for the blocks in height range.
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode
        }
        """

        start_height = int(start_height)
        end_height = int(end_height)

        if start_height < 0:
            raise ValueError("start_height < 0")
        if end_height < start_height:
            raise ValueError("end_height < start_height")

        return self.raw_jsonrpc_request(
            "get_block_headers_range",
            params={"start_height": start_height, "end_height": end_height},
        )

    def get_block(self, height=None, hash=None):
        """
        Full block information can be retrieved by either block height or hash. Either height OR
        hash must be specified, but not both.

        :param int height: block height
        :param str hash: block hash

        Output:
        {
        "blob": str; Hexadecimal blob of block information.
        "block_header": A structure containing block header information. See get_last_block_header.
        "json": str; JSON formatted string details:
            "major_version": unsigned int; The major version of the monero protocol at this block height.
            "minor_version": unsigned int; The minor version of the monero protocol at this block height.
            "timestamp": unsigned int; The unix time at which the block was recorded into the blockchain.
            "prev_id": str; The hash of the block immediately preceding this block in the chain.
            "nonce": unsigned int; a cryptographic random one-time number used in mining a Monero block.
            "miner_tx": {
                "version": unsigned int; Transaction version number.
                "unlock_time": unsigned int; The block height when the coinbase transaction becomes spendable.
                "vin": list; List of transaction inputs with the following structure:
                    {
                        "gen": {        (note: miner txs are coinbase txs or "gen")
                            "height": This block height, a.k.a. when the coinbase is generated.
                        }
                    }
                "vout": list; List of transaction outputs with the following structures:
                    {
                        "amount"; unsigned int; The amount of the output, in atomic units.
                        "target": {
                            "key": str; public key of one-time output of miner tx.
                        }
                    }
                "extra"; list; List of integers 0-255, usually the "transaction ID" but can be used to include byte string.
                "rct_signatures" - Contain signatures of tx signers. Coinbased txs do not have signatures.
        "miner_tx_hash": str; hash of the coinbase transaction which belongs to this block.
        "tx_hashes": list; List of str hex hashes of non-coinbase transactions in the block.
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        if height is None and hash is None:
            raise ValueError("height or hash must be provided")
        elif height is not None and hash is not None:
            raise ValueError("height and hash can not both be provided")

        if height is not None:
            if height < 0:
                raise ValueError("{} is not a valid height".format(height))

            params = {"height": int(height)}
        else:
            if not self._is_valid_256_hex(hash):
                raise ValueError('blk_hash "{}" is not a valid hash'.format(hash))

            params = {"hash": hash}

        return self.raw_jsonrpc_request("get_block", params=params)

    def get_connections(self):
        """
        Retrieve information about incoming and outgoing connections to your node.

        Output:
        {
        "connections": list; List of all connections and their info with the following structure:
            {
            "address": str; The peer's address, actually IPv4 & port
            "avg_download": unsigned int; Average bytes of data downloaded by node.
            "avg_upload": unsigned int; Average bytes of data uploaded by node.
            "connection_id": str; The connection ID
            "current_download": unsigned int; Current bytes downloaded by node.
            "current_upload": unsigned int; Current bytes uploaded by node.
            "height": unsigned int; The peer height
            "host": str; The peer host
            "incoming": bool; Is the node getting information from your node?
            "ip": str; The node's IP address.
            "live_time": unsigned int
            "local_ip": bool
            "localhost": bool
            "peer_id": str; The node's ID on the network.
            "port": str; The port that the node is using to connect to the network.
            "recv_count": unsigned int
            "recv_idle_time": unsigned int
            "send_count": unsigned int
            "send_idle_time": unsigned int
            "state": str
            "support_flags": unsigned int
            }
        }
        """

        return self.raw_jsonrpc_request("get_connections")

    def get_info(self):
        """
        Retrieve general information about the state of your node and the network.

        Output:
        {
        "alt_blocks_count": unsigned int; Number of alternative blocks to main chain.
        "block_size_limit": unsigned int; Maximum allowed block size
        "block_size_median": unsigned int; Median block size of latest 100 blocks
        "bootstrap_daemon_address": str; bootstrap node to give immediate usability to wallets while syncing by proxying RPC to it.
        "busy_syncing": bool; States if new blocks are being added (true) or not (false).
        "cumulative_difficulty": unsigned int; Cumulative difficulty of all blocks in the blockchain.
        "difficulty": unsigned int; Network difficulty (analogous to the strength of the network)
        "free_space": unsigned int; Available disk space on the node.
        "grey_peerlist_size": unsigned int; Grey Peerlist Size
        "height": unsigned int; Current length of longest chain known to daemon.
        "height_without_bootstrap": unsigned int; Current length of the local chain of the daemon.
        "incoming_connections_count": unsigned int; Number of peers connected to and pulling from your node.
        "mainnet": bool; States if the node is on the mainnet (true) or not (false).
        "offline": bool; States if the node is offline (true) or online (false).
        "outgoing_connections_count": unsigned int; Number of peers that you are connected to and getting information from.
        "rpc_connections_count": unsigned int; Number of RPC client connected to the daemon (Including this RPC request).
        "stagenet": bool; States if the node is on the stagenet (true) or not (false).
        "start_time": unsigned int; Start time of the daemon, as UNIX time.
        "status": str; General RPC error code. "OK" means everything looks good.
        "synchronized": bool; States if the node is synchronized (true) or not (false).
        "target": unsigned int; Current target for next proof of work.
        "target_height": unsigned int; The height of the next block in the chain.
        "testnet": bool; States if the node is on the testnet (true) or not (false).
        "top_block_hash": str; Hash of the highest block in the chain.
        "tx_count": unsigned int; Total number of non-coinbase transaction in the chain.
        "tx_pool_size": unsigned int; Number of transactions that have been broadcast but not included in a block.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        "was_bootstrap_ever_used": bool; States if a bootstrap node has ever been used since the daemon started.
        "white_peerlist_size": unsigned int; White Peerlist Size
        }
        """

        return self.raw_jsonrpc_request("get_info")

    def hard_fork_info(self):
        """
        Look up information regarding hard fork voting and readiness.

        Output:
        {
        "earliest_height": unsigned int; Block height at which hard fork would be enabled if voted in.
        "enabled": bool; Tells if hard fork is enforced.
        "state": unsigned int; Current hard fork state: 0 (There is likely a hard fork), 1 (An update is needed to fork properly), or 2 (Everything looks good).
        "status": str; General RPC error code. "OK" means everything looks good.
        "threshold": unsigned int; Minimum percent of votes to trigger hard fork. Default is 80.
        "version": unsigned int; The major block version for the fork.
        "votes": unsigned int; Number of votes towards hard fork.
        "voting": unsigned int; Hard fork voting status.
        "window": unsigned int; Number of blocks over which current votes are cast. Default is 10080 blocks.
        }
        """

        return self.raw_jsonrpc_request("hard_fork_info")

    def set_bans(self, ip, ban, seconds=None):
        """
        Ban (or unban) other nodes by IP for a certain length of time.

        :param ip: ip address of node to ban in two forms: 'a.b.c.d' str or 32-bit unsigned integer
        :param bool ban: Whether to ban (True) or unban (False)
        :param int seconds: Number of seconds to ban. Optional if "ban" is False

        Notes:
        * Params "ip", "ban", and "seconds" can be single elements or they can be iterables of the same length. For example:

            ips =        ['229.52.8.67', 611282986]
            should_ban = [True,          False]
            seconds =    [10000,         None]
            daemon.set_bans(ips, should_ban, seconds)

        * "seconds" can be 0 or None if "ban" is False

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        if isinstance(ip, (str, int)):  # If single element paramters
            user_bans = [{"ip": ip, "ban": ban, "seconds": seconds}]
        else:  # If params are iterables, contrust the list of ban entries
            if not (len(ip) == len(ban) == len(seconds)):
                raise ValueError('"ip", "ban", and "seconds" are not the same length')

            user_bans = [
                {"ip": ip[i], "ban": ban[i], "seconds": seconds[i]}
                for i in range(len(ip))
            ]

        # Construct the bans parameter to be passed to the RPC command
        bans = []
        for ban_entry in user_bans:
            curr_ip = ban_entry["ip"]
            curr_ban = bool(ban_entry["ban"])
            curr_seconds = ban_entry["seconds"]

            # validate IP
            ipaddress.IPv4Address(curr_ip)

            if curr_ban and curr_seconds is None:
                raise ValueError('Whenever "ban" is True, "seconds" must not be None')
            elif curr_seconds is not None and curr_seconds < 0:
                raise ValueError('"seconds" can not be less than zero')

            ban = {}
            if isinstance(curr_ip, int):
                ban["ip"] = curr_ip
            else:
                ban["host"] = curr_ip
            if curr_seconds is not None:
                ban["seconds"] = curr_seconds
            ban["ban"] = curr_ban

            bans.append(ban)

        return self.raw_jsonrpc_request("set_bans", params={"bans": bans})

    def get_bans(self):
        """
        Get list of ban entries. For structure of ban entires, see documentation of set_bans method.

        Output:
        {
        "bans": list; banned nodes with following structure
            {
                "host": str; Banned host (IP in A.B.C.D form).
                "ip": unsigned int; Banned IP address, in Int format.
                "seconds": unsigned int; Local Unix time that IP is banned until.
            }
        "status": str; General RPC error code. "OK" means everything looks good.
        }
        """

        resp = self.raw_jsonrpc_request("get_bans")

        # When there are no ban entries, the node returns a responses with no "bans" field.
        # This is counterintuitive for the user, so I add an empty field if it's not provided.
        if "bans" not in resp:
            resp["bans"] = []

        return resp

    def flush_txpool(self, txids=None):
        """
        Flush tx ids from transaction pool

        :param list txids: list of str; transactions IDs to flush from mempool (all tx ids flushed if empty).

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        txids = self._validate_hashlist(txids)

        resp = self.raw_jsonrpc_request("flush_txpool", params={"txids": txids})

        if "transactions" not in resp:
            resp["transactions"] = []

        return resp

    def get_output_histogram(
        self, amounts, min_count=None, max_count=None, unlocked=None, recent_cutoff=None
    ):
        """
        Get a histogram of output amounts. For all amounts (possibly filtered by parameters), gives the number
        of outputs on the chain for that amount. RingCT outputs counts as 0 amount.

        :param list amounts: list of unsigned ints in atomic units, or Decimals as full monero amounts
        :param int min_count: lower bound for output number
        :param int max_count: upper bound for output number
        :param bool unlocked: whether to count only unlocked
        :pramrm int recent_cutoff: unsigned int


        Output:
        {
        "histogram": list; histogram entries with the following structure
            {
            "amount": unsigned int; Output amount in atomic units.
            "total_instances": unsigned int; number of total instances for given amount.
            "unlocked_instances": unsigned int; number of unlocked instances for given amount.
            "recent_instances": unsigned int; number of recent instances for given amount.
            }
        "status - string; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        # Coerce amounts paramter
        if isinstance(amounts, (Decimal, int)):
            amounts = [to_atomic(amounts) if isinstance(amounts, Decimal) else amounts]
        elif amounts is None:
            raise ValueError("amounts is None")

        # Construct RPC parameters
        params = {"amounts": amounts}
        if min_count is not None:
            params["min_count"] = int(min_count)
        if max_count is not None:
            params["max_count"] = int(max_count)
        if unlocked is not None:
            params["unlocked"] = bool(unlocked)
        if recent_cutoff is not None:
            params["recent_cutoff"] = int(recent_cutoff)

        return self.raw_jsonrpc_request("get_output_histogram", params=params)

    def get_coinbase_tx_sum(self, height, count):
        """
        Get the coinbase amount and the fees amount for n last blocks starting at particular height.

        :param int height: unsigned int; block height from which getting the amounts
        :param int count: unsigned int; number of blocks to include in the sum

        Output:
        {
        "emission_amount": unsigned int; amount of coinbase reward in atomic units
        "fee_amount": unsigned int; amount of fees in atomic units
        "status": str; General RPC error code. "OK" means everything looks good.
        }
        """

        if height < 0:
            raise ValueError("height < 0")
        elif count < 0:
            raise ValueError("count < 0")

        return self.raw_jsonrpc_request(
            "get_coinbase_tx_sum", params={"height": height, "count": count}
        )

    def get_version(self):
        """
        Get the node's current version.

        Output:
        {
        "version": unsigned int;
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """
        return self.raw_jsonrpc_request("get_version")

    def get_fee_estimate(self, grace_blocks=None):
        """
        Gives an estimation on fees per byte.

        :param int grace_blocks: Optional; number of previous blocks to include in fee calculation

        Output:
        {
        "fee": unsigned int; Amount of fees estimated per byte in atomic units
        "quantization_mask": unsigned int; Final fee should be rounded up to an even multiple of this value
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        if not isinstance(grace_blocks, (type(None), int)):
            raise TypeError("grace_blocks is not an int")
        elif grace_blocks is not None and grace_blocks < 0:
            raise ValueError("grace_blocks < 0")

        params = {"grace_blocks": grace_blocks} if grace_blocks else None
        return self.raw_jsonrpc_request("get_fee_estimate", params=params)

    def get_alternate_chains(self):
        """
        Display alternative chains seen by the node.


        Output:
        {
        "chains": list; chain informations with the following structure
            {
                "block_hash": str; the block hash of the first diverging block of this alternative chain.
                "difficulty": unsigned int; the cumulative difficulty of all blocks in the alternative chain.
                "height": unsigned int; the block height of the first diverging block of this alternative chain.
                "length": unsigned int; the length in blocks of this alternative chain, after divergence.
            }
        "status": str; General RPC error code. "OK" means everything looks good.
        }

        """

        return self.raw_jsonrpc_request("get_alternate_chains")

    def relay_tx(self, txids):
        """
        Relay a list of transaction IDs.

        :param list txids: list of transaction IDs to relay

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        txids = self._validate_hashlist(txids)

        return self.raw_jsonrpc_request("relay_tx", params={"txids": txids})

    def sync_info(self):
        """
        Get synchronisation information.

        Output:
        {
        "height": unsigned int; current height
        "peers": list; peer informations with the following structure
            {
                "info": dict; structure of connection info, as defined in get_connections
            }
        "spans": list; span informations with the follwing structure (absent if node is fully synced)
            {
                "connection_id": str; connection ID.
                "nblocks": unsigned int; number of blocks in that span.
                "rate": unsigned int; connection rate.
                "remote_address": str; peer address the node is downloading (or has downloaded) the span from.
                "size": unsigned int; total number of bytes in that span's blocks (including txes).
                "speed": unsigned int; connection speed.
                "start_block_height": unsigned int; block height of the first block in that span.
            }
        "status": str; General RPC error code. "OK" means everything looks good.
        "target_height": unsigned int; target height the node is syncing from (will be undefined if node is fully synced)
        }
        """

        return self.raw_jsonrpc_request("sync_info")

    # Other RPC Methods (https://www.getmonero.org/resources/developer-guides/daemon-rpc.html#other-daemon-rpc-calls)

    def get_height(self):
        """
        Get the node's current height.

        Outputs:
        {
        "height": unsigned int; Current length of longest chain known to daemon.
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        return self.raw_request("/get_height")

    def get_transactions(self, tx_hashes, decode_as_json=False, prune=False):
        """
        Look up one or more transactions by hash.

        :param list txs_hashes: List of transaction hashes to look up.
        :param bool decode_as_json: Optional (false by default). If true, the returned tx info will be decoded into JSON.
        :param bool prune: Optional (false by default). If true, prune blob data, greatly cutting down on response size.


        Output:
        {
        "missed_tx": list; (Optional - returned if not empty) Transaction hashes that could not be found.
        "status": str; General RPC error code. "OK" means everything looks good.
        "txs": list; transaction informations with the following structure
            {
                "as_hex": stri; Full transaction information as a hex string.
                "as_json": json str; JSON formatted as follows
                    {
                        "version": unsigned int; Transaction version
                        "unlock_time": unsigned int; If not 0, this tells when a transaction output is spendable.
                        "vin": list; transaction inputs with the following structure
                            {
                                "key":
                                {
                                    "amount": unsigned int; The amount of the input, in atomic units.
                                    "key_offsets": list; integer offets to the input.
                                    "k_image": str; The key image for the given input
                                }

                            }
                        "vout": list; transaction inputs with the following structure
                            {
                                "amount": unsigned int; Amount of transaction output, in atomic units.
                                "target":
                                {
                                    "key": str; The stealth public key of the receiver.
                                }
                            }
                        "extra"; list; List of integers 0-255, usually the "transaction ID" but can be used to include byte string.
                        signatures - List of signatures used in ring signature to hide the true origin of the transaction.
                    }
                "block_height": unsigned int; block height including the transaction
                "block_timestamp": unsigned int; Unix time at chich the block has been added to the blockchain
                "double_spend_seen": boolean; States if the transaction is a double-spend (true) or not (false)
                "in_pool": bool; States if the transaction is in pool (true) or included in a block (false)
                "output_indices": list; list of the tx's output's globals indexes as unsigned ints
                "tx_hash": str; transaction hash
            }
        "txs_as_hex": str; Full transaction information as a hex string (old compatibility parameter)
        "txs_as_json": json str; (Optional - returned if set in inputs. Old compatibility parameter) List of transaction as in as_json above:
        }
        """

        tx_hashes = self._validate_hashlist(tx_hashes)

        return self.raw_request(
            "/get_transactions",
            data={
                "txs_hashes": tx_hashes,
                "decode_as_json": bool(decode_as_json),
                "prune": bool(prune),
            },
        )

    def get_alt_blocks_hashes(self):
        """
        Get the known blocks hashes which are not on the main chain.

        Output:
        {
        "blks_hashes"; list; alternative blocks hashes to main chain
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        return self.raw_request("/get_alt_blocks_hashes")

    def is_key_image_spent(self, key_images):
        """
        Check if outputs have been spent using the key image associated with the output.

        :param list key_images: List of key image hex strings to check.

        Key Image Status Legend:
            0 = unspent
            1 = spent in blockchain
            2 = spent in transaction pool

        Outputs:
        {
        "spent_status": list of unsigned ints; List of statuses for each image checked.
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        key_images = self._validate_hashlist(key_images)

        return self.raw_request("/is_key_image_spent", data={"key_images": key_images})

    def send_raw_transaction(self, tx_as_hex, do_not_relay=False):
        """
        Broadcast a raw transaction to the network.

        :param str tx_as_hex: Full transaction information as hexidecimal string.
        :param bool do_not_relay: Optional; Stop relaying transaction to other nodes (default is false).

        Output:
        {
        "double_spend": bool; Transaction is a double spend (true) or not (false).
        "fee_too_low": bool; Fee is too low (true) or OK (false).
        "invalid_input": bool; Input is invalid (true) or valid (false).
        "invalid_output": bool; Output is invalid (true) or valid (false).
        "low_mixin": bool; Mixin count is too low (true) or OK (false).
        "not_rct": bool; Transaction is a standard ring transaction (true) or a ring confidential transaction (false).
        "not_relayed": bool; Transaction was not relayed (true) or relayed (false).
        "overspend": bool; Transaction uses more money than available (true) or not (false).
        "reason": str; Additional information. Currently empty or "Not relayed" if transaction was accepted but not relayed.
        "too_big": bool; Transaction size is too big (true) or OK (false).
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """
        return self.raw_request(
            "/send_raw_transaction",
            data={"tx_as_hex": tx_as_hex, "do_not_relay": do_not_relay},
        )

    def start_mining(
        self, do_background_mining, ignore_battery, miner_address, threads_count
    ):
        """
        Start mining on the daemon.

        :param bool do_background_mining: States if the mining should run in background (true) or foreground (false).
        :param bool ignore_battery: States if batery state (on laptop) should be ignored (true) or not (false).
        :param str miner_address: Account address to mine to.
        :param int threads_count: Number of mining thread to run.

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        if isinstance(miner_address, address.Address):
            miner_address = repr(miner_address)
        else:
            try:
                converted_addr = address.address(miner_address)
            except ValueError:
                raise ValueError(
                    'miner_address "{}" does not match any recognized address format'.format(
                        miner_address
                    )
                )

            if not isinstance(converted_addr, address.Address):
                raise ValueError(
                    'miner_address must be a "standard" monero address (i.e. it must start with a "4")'
                )

        if not isinstance(threads_count, int):
            raise TypeError("threads_count must be an int")

        if threads_count < 0:
            raise ValueError("threads_count < 0")

        return self.raw_request(
            "/start_mining",
            data={
                "do_background_mining": bool(do_background_mining),
                "ignore_battery": bool(ignore_battery),
                "miner_address": miner_address,
                "threads_count": threads_count,
            },
        )

    def stop_mining(self):
        """
        Stop mining on the daemon.

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        return self.raw_request("/stop_mining")

    def mining_status(self):
        """
        Get the mining status of the daemon.

        Output:
        {
        "active": bool; States if mining is enabled (true) or disabled (false).
        "address": str; Account address daemon is mining to. Empty if not mining.
        "is_background_mining_enabled": bool; States if the mining is running in background (true) or foreground (false).
        "speed": unsigned int; Mining power in hashes per seconds.
        "threads_count": unsigned int; Number of running mining threads.
        "status":  str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        return self.raw_request("/mining_status")

    def save_bc(self):
        """
        Save the blockchain. The blockchain does not need saving and is always saved when modified, however it does a sync to
        flush the filesystem cache onto the disk for safety purposes against Operating System or Harware crashes.

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        return self.raw_request("/save_bc")

    def get_peer_list(self):
        """
        Get the known peer list.

        Outputs:
        {
        "gray_list": list; offline peer informations with the following structure
            {
                "host": unsigned int; IP address in integer format.
                "id": stri; Peer id.
                "ip": unsigned int; IP address in integer format.
                "last_seen": unsigned int; unix time at which the peer has been seen for the last time.
                "port": unsigned int; TCP port the peer is using to connect to monero network.
            }
        "white_list": list; online peer informations with the above structure ^^^
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        return self.raw_request("/get_peer_list")

    def set_log_hash_rate(self, visible):
        """
        Set the log hash rate display mode.

        :param bool visible: States if hash rate logs should be visible (true) or hidden (false)

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        resp = self.raw_request("/set_log_hash_rate", data={"visible": bool(visible)})

        if resp["status"] == "NOT MINING":
            raise RPCError(
                'The node at "{url}" is not currently mining and therefore cannot set its hash rate log visibility.'.format(
                    url=self.url
                )
            )

        return resp

    def set_log_level(self, level):
        """
        Set the daemon log level. By default, log level is set to 0.

        :param int level: daemon log level to set from 0 (less verbose) to 4 (most verbose)

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        if level not in range(5):
            raise ValueError(
                "level must an integer between 0 (less verbose) and 4 (more verbose), inclusive"
            )

        return self.raw_request("/set_log_level", data={"level": level})

    def set_log_categories(self, categories=None):
        """
        Set the daemon log categories. Categories are represented as a comma separated list of
        <Category>:<level> (similarly to syslog standard <Facility>:<Severity-level>).
        To get a list of all log categories known by this library, call JSONRPCDaemon.get_all_known_log_categories().
        To get a list of all log levels known by this library, call JSONRPCDaemon.get_all_known_log_levels().

        :param categories: str or list; Optional, daemon log categories to enable. Accepts list or comma-seperated string.

        Output:
        {
        "categories": str; comma-seperated string list of daemon log enabled categories
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        if not categories:
            return self.raw_request("/set_log_categories")
        elif isinstance(categories, str):
            categories = categories.split(",")

        # Validate categories
        for cat_str in categories:
            try:
                category, level = cat_str.split(":")

                if category not in self._KNOWN_LOG_CATEGORIES:
                    _log.warn('Unrecognized log category: "{}"'.format(category))

                if level not in self._KNOWN_LOG_LEVELS:
                    _log.warn('Unrecognized log level: "{}"'.format(level))
            except ValueError:
                raise ValueError('malformed category string: "{}"'.format(cat_str))

        final_category_str = ",".join(categories)

        return self.raw_request(
            "/set_log_categories", data={"categories": final_category_str}
        )

    def get_transaction_pool(self):
        """
        Show information about valid transactions seen by the node but not yet mined into a block,
        as well as spent key image information for the txpool in the node's memory.

        Output:
        {
        "spent_key_images": list; spent output key images with following structure
            {
                "id_hash": str; Key image.
                "txs_hashes": list; tx hashes of the txes (usually one) spending that key image.
            }
        "transactions": list; transactions in the mempool with the following structure
            {
                "blob_size": unsigned int; The size of the full transaction blob.
                "double_spend_seen": bool; States if this transaction has been seen as double spend.
                "do_not_relay": bool; States if this transaction should not be relayed
                "fee": unsigned int; The amount of the mining fee included in the transaction, in atomic units.
                "id_hash": str; The transaction ID hash.
                "kept_by_block": bool; States if the tx was included in a block at least once (true) or not (false).
                "last_failed_height": unsigned int; If the transaction validation has previously failed, this tells at what height that occured.
                "last_failed_id_hash": str; Like the previous, this tells the previous transaction ID hash.
                "last_relayed_time": unsigned int; Last unix time at which the transaction has been relayed.
                "max_used_block_height": unsigned int; Tells the height of the most recent block with an output used in this transaction.
                "max_used_block_hash": str; Tells the hash of the most recent block with an output used in this transaction.
                "receive_time": unsigned int; The Unix time that the transaction was first seen on the network by the node.
                "relayed": bool; States if this transaction has been relayed
                "tx_blob": unsigned int; Hexadecimal blob represnting the transaction.
                "tx_json": json string; JSON structure of all information in the transaction (see get_transactions() for structure information)
            }
        "status": str; General RPC error code. "OK" means everything looks good.
        }
        """

        return self.raw_request("/get_transaction_pool")

    def get_transaction_pool_stats(self):
        """
        Get the transaction pool statistics.

        Output:
        {
        "pool_stats": {
            "bytes_max": unsigned int; Max transaction size in pool.
            "bytes_med" unsigned int; Median transaction size in pool.
            "bytes_min" unsigned int; Min transaction size in pool.
            "bytes_total": unsigned int; total size of all transactions in pool.
            "histo": list; histogram of transaction sizes with the following structure
                {
                    "txs": unsigned int; number of transactions.
                    "bytes": unsigned int; size in bytes.
                }
            "histo_98pc": unsigned int; the time 98% of txes are "younger" than.
            "num_10m": unsigned int; number of transactions in pool for more than 10 minutes.
            "num_double_spends": unsigned int; number of double spend transactions.
            "num_failing": unsigned int; number of failing transactions.
            "num_not_relayed": unsigned int; number of non-relayed transactions.
            "oldest": unsigned int; unix time of the oldest transaction in the pool
            "txs_total": unsigned int; total number of transactions.
        }
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        return self.raw_request("/get_transaction_pool_stats")

    def stop_daemon(self):
        """
        Send a command to the daemon to safely disconnect and shut down.

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        return self.raw_request("/stop_daemon")

    def get_limit(self):
        """
        Get daemon bandwidth limits.

        Outputs:
        {
        "limit_down": unsigned int; Download limit in kBytes per second
        "limit_up": unsigned int; Upload limit in kBytes per second
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        return self.raw_request("/get_limit")

    def set_limit(self, limit_down=-1, limit_up=-1):
        """
        Set daemon bandwidth limits.

        :param int limit_down: Download limit in kBytes per second (-1 reset to default, 0 don't change the current limit)
        :param int limit_up: Upload limit in kBytes per second (-1 reset to default, 0 don't change the current limit)

        Output:
        {
        "limit_down": - unsigned int; Download limit in kBytes per second
        "limit_up": unsigned int; Upload limit in kBytes per second
        "status": str; General RPC error code. "OK" means everything looks good.
        }
        """

        if limit_down < -1:
            raise ValueError("invalid limit_down: {}".format(limit_down))
        elif limit_up < -1:
            raise ValueError("invalid limit_up: {}".format(limit_up))

        return self.raw_request(
            "/set_limit", data={"limit_down": limit_down, "limit_up": limit_up}
        )

    def out_peers(self, out_peers_arg):
        """
        Limit number of outgoing peers.

        :param int out_peers_arg: Max number of outgoing peers. If less than zero, allow unlimited outgoing peers

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        if out_peers_arg < 0:
            out_peers_arg = 2**32 - 1
        elif out_peers_arg > 2**32 - 1:
            raise ValueError('out_peers_arg "{}" is too large'.format(out_peers_arg))

        return self.raw_request("/out_peers", data={"out_peers": int(out_peers_arg)})

    def in_peers(self, in_peers_arg):
        """
        Limit number of incoming peers.

        :param int in_peers_arg: Max number of incoming peers. If less than zero, allow unlimited incoming peers

        Output:
        {
        "status": str; General RPC error code. "OK" means everything looks good. Any other value means that something went wrong.
        }
        """

        if in_peers_arg < 0:
            in_peers_arg = 2**32 - 1
        elif in_peers_arg > 2**32 - 1:
            raise ValueError('in_peers_arg "{}" is too large'.format(in_peers_arg))

        return self.raw_request("/in_peers", data={"in_peers": int(in_peers_arg)})

    def get_outs(self, amount, index, get_txid=True):
        """
        Get information about one-time outputs (stealth addresses).


        :param amount: list or single element of output amount. Decimal for full monero amounts or int for atmoic units
        :param index: list of single element of output index. int
        :param bool get_txid: Optional. If true, a txid will included for each output in the response.

        Output:
        {
        "outs": list; output informations with the following structure
            {
                "height": unsigned int; block height of the output
                "key": str; the public key of the output
                "mask": str;
                "txid": str; transaction id
                "unlocked": bool; States if output is locked (false) or not (true)
            }
        "status": str; General RPC error code. "OK" means everything looks good.
        "untrusted": bool; True for bootstrap mode, False for full sync mode.
        }
        """

        if isinstance(index, int):  # single element
            outputs = [
                {
                    "amount": amount,
                    "index": index,
                }
            ]
        else:  # multiple elements
            if len(amount) != len(index):
                raise ValueError("length of amount and index do not match")

            outputs = []
            for a, i in zip(amount, index):
                outputs.append(
                    {
                        "amount": a if isinstance(a, int) else to_atomic(a),
                        "index": i,
                    }
                )

        return self.raw_request(
            "/get_outs", data={"outputs": outputs, "get_txid": get_txid}
        )

    def update(self, command, path=None):
        """
        Update daemon.


        :param str command: Command to use, either "check" or "download".
        :param str path: Optional, path where to download the update.


        Output:
        {
        "auto_uri": str;
        "hash": str;
        "path": str; path to download the update.
        "update": bool; States if an update is available to download (true) or not (false).
        "user_uri": str;
        "version": str; Version available for download.
        "status": str; General RPC error code. "OK" means everything looks good.
        }
        """
        if command not in ("check", "download"):
            raise ValueError('unrecognized command: "{}"'.format(command))

        data = {"command": command}

        if path is not None:
            data["path"] = path

        return self.raw_request("/update", data=data)

    # Supporting class methods

    @classmethod
    def known_log_categories(cls):
        return cls._KNOWN_LOG_CATEGORIES

    @classmethod
    def known_log_levels(cls):
        return cls._KNOWN_LOG_LEVELS

    # private methods

    def _set_net(self, info):
        if info["mainnet"]:
            self._net = NET_MAIN
        if info["testnet"]:
            self._net = NET_TEST
        if info["stagenet"]:
            self._net = NET_STAGE

    def _set_restricted(self):
        try:
            self.sync_info()

            self._restricted = False
        except MethodNotFound:
            self._restricted = True

    def _do_get_transactions(self, hashes, prune):
        res = self.raw_request(
            "/get_transactions",
            {"txs_hashes": hashes, "decode_as_json": True, "prune": prune},
        )
        if res["status"] != "OK":
            raise exceptions.BackendException(res["status"])
        txs = []
        for tx in res.get("txs", []):
            as_json = json.loads(tx["as_json"])
            fee = as_json.get("rct_signatures", {}).get("txnFee")
            txs.append(
                Transaction(
                    hash=tx["tx_hash"],
                    fee=from_atomic(fee) if fee else None,
                    height=None if tx["in_pool"] else tx["block_height"],
                    timestamp=datetime.fromtimestamp(tx["block_timestamp"])
                    if "block_timestamp" in tx
                    else None,
                    output_indices=tx["output_indices"]
                    if "output_indices" in tx
                    else None,
                    blob=binascii.unhexlify(tx["as_hex"]) or None,
                    json=as_json,
                )
            )

        return txs

    def _is_valid_256_hex(self, hexhash):
        try:
            bytearray.fromhex(hexhash)

            return len(hexhash) == 64
        except ValueError:
            return False

    def _validate_hashlist(self, hashes):
        if not hashes:
            return []
        else:
            if isinstance(hashes, str):
                hashes = [hashes]
            else:
                coerce_compatible_str = (
                    lambda s: s.decode() if isinstance(s, bytes) else s
                )
                hashes = list(map(coerce_compatible_str, hashes))

            not_valid = lambda h: not self._is_valid_256_hex(h)

            if any(map(not_valid, hashes)):
                raise ValueError("hashes contains an invalid hash")

            return hashes
